from types import TracebackType
from typing import Callable, Protocol, runtime_checkable, TypeVar

from omnipy.api.typedefs import DecoratorClassT
from omnipy.util.setdeque import SetDeque

_ObjT = TypeVar('_ObjT', bound=object)
_ObjContraT = TypeVar('_ObjContraT', contravariant=True, bound=object)
_AnyKeyT = TypeVar('_AnyKeyT', contravariant=True, bound=object)
_ValT = TypeVar('_ValT', bound=object)
_ContentsT = TypeVar('_ContentsT', bound=object)
_ContentCovT = TypeVar('_ContentCovT', covariant=True, bound=object)
_ContentContraT = TypeVar('_ContentContraT', contravariant=True, bound=object)
_HasContentsT = TypeVar('_HasContentsT', bound='HasContents')


@runtime_checkable
class IsCallableParamAfterSelf(Protocol):
    """"""
    def __call__(self, callable_arg: Callable, /, *args: object, **kwargs: object) -> None:
        ...


@runtime_checkable
class IsCallableClass(Protocol[DecoratorClassT]):
    """"""
    def __call__(self, *args: object, **kwargs: object) -> Callable[[Callable], DecoratorClassT]:
        ...


@runtime_checkable
class HasContents(Protocol[_ContentsT]):
    @property
    def contents(self) -> _ContentsT:
        ...

    @contents.setter
    def contents(self, value: _ContentsT) -> None:
        ...


class IsWeakKeyRefContainer(Protocol[_AnyKeyT, _ValT]):
    def __contains__(self, key: _AnyKeyT) -> bool:
        ...

    def get(self, key: _AnyKeyT) -> _ValT | None:
        ...

    def __getitem__(self, key: _AnyKeyT) -> _ValT:
        ...

    def __setitem__(self, key: _AnyKeyT, value: _ValT) -> None:
        ...

    def __len__(self) -> int:
        ...


class IsSnapshotWrapper(Protocol[_ObjContraT, _ContentsT]):
    id: int
    snapshot: _ContentsT

    def taken_of_same_obj(self, obj: _ObjContraT) -> bool:
        ...

    def differs_from(self, obj: _ObjContraT) -> bool:
        ...


class IsSnapshotHolder(IsWeakKeyRefContainer[_HasContentsT,
                                             IsSnapshotWrapper[_HasContentsT, _ContentsT]],
                       Protocol[_HasContentsT, _ContentsT]):
    """"""
    def clear(self) -> None:
        ...

    def all_are_empty(self, debug: bool = False) -> bool:
        ...

    def get_deepcopy_content_ids(self) -> SetDeque[int]:
        ...

    def get_deepcopy_content_ids_scheduled_for_deletion(self) -> SetDeque[int]:
        ...

    def schedule_deepcopy_content_ids_for_deletion(self, *keys: int) -> None:
        ...

    def delete_scheduled_deepcopy_content_ids(self) -> None:
        ...

    def take_snapshot_setup(self) -> None:
        ...

    def take_snapshot_teardown(self) -> None:
        ...

    def take_snapshot(self, obj: _HasContentsT) -> None:
        ...


class IsRateLimitingClientSession(Protocol):
    """"""
    @property
    def requests_per_second(self) -> float:
        ...

    async def __aenter__(self) -> 'IsRateLimitingClientSession':
        ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        ...

    async def close(self) -> None:
        ...



from io import StringIO
from typing import Mapping, Sequence

from omnipy.compute.task import TaskTemplate
from omnipy.compute.typing import mypy_fix_task_template
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

from . import pd
from ..general.models import NotIterableExceptStrOrBytesModel
from .helpers import extract_common_colnames
from .models import ListOfPandasDatasetsWithSameNumberOfFiles, PandasDataset, PandasModel


@mypy_fix_task_template
@TaskTemplate()
def convert_dataset_list_of_dicts_to_pandas(
        dataset: Dataset[Model[list[dict[str, NotIterableExceptStrOrBytesModel]]]]) \
        -> PandasDataset:
    pandas_dataset = PandasDataset()
    pandas_dataset.from_data(dataset.to_data())
    return pandas_dataset


@mypy_fix_task_template
@TaskTemplate()
def convert_dataset_csv_to_pandas(dataset: Dataset[Model[bytes]],
                                  delimiter: str = ',',
                                  first_row_as_col_names=True,
                                  col_names: list[str] | None = None,
                                  ignore_comments: bool = True,
                                  comments_char: str = '#') -> PandasDataset:
    out_dataset = PandasDataset()
    for key, item in dataset.items():
        df = pd.read_csv(
            StringIO(item.contents),
            sep=delimiter,
            header='infer' if first_row_as_col_names else 0,
            names=col_names,
            comment=comments_char if ignore_comments else None,
            encoding='utf8',
        )
        out_dataset[key] = df
    return out_dataset


@mypy_fix_task_template
@TaskTemplate()
def convert_dataset_pandas_to_csv(
    dataset: PandasDataset,
    delimiter: str = ',',
    first_row_as_col_names=True,
    col_names: list[str] | None = None,
) -> Dataset[Model[str]]:
    out_dataset = Dataset[Model[str]]()
    for key, df in dataset.items():
        csv_stream = StringIO()
        df.to_csv(
            csv_stream,
            sep=delimiter,
            header=col_names if col_names else True if first_row_as_col_names else False,
            encoding='utf8',
            index=False)
        out_dataset[key] = csv_stream.getvalue()
    return out_dataset


@mypy_fix_task_template
@TaskTemplate()
def extract_columns_as_files(dataset: PandasDataset, col_names: list[str]) -> PandasDataset:
    out_dataset = PandasDataset()
    for key, item in dataset.items():
        df = dataset[key]
        out_dataset[key] = df.loc[:, ~df.columns.isin(col_names)]

        for col_name in col_names:
            out_dataset[f'{key}.{col_name}'] = pd.DataFrame(df[col_name])
    return out_dataset


@mypy_fix_task_template
@TaskTemplate()
def concat_dataframes_across_datasets(dataset_list: ListOfPandasDatasetsWithSameNumberOfFiles,
                                      vertical=True) -> PandasDataset:
    # We know from the data type that there are at least two datasets and that there is an equal
    # amount of files/DataFrames in each dataset. This simplifies implementation.

    out_dataset = PandasDataset()
    out_datafile_names = tuple(dataset_list[0].keys())
    for df_index in range(len(out_datafile_names)):
        df = pd.concat([tuple(dataset.values())[df_index] for dataset in dataset_list],
                       axis=0 if vertical else 1)
        out_dataset[out_datafile_names[df_index]] = df
    return out_dataset


@mypy_fix_task_template
@TaskTemplate()
def join_tables(table_1: PandasModel,
                table_2: PandasModel,
                join_type: str = 'outer',
                on_cols: Sequence[str] | Mapping[str, str] | None = None) -> PandasModel:
    if join_type == 'cross':
        raise ValueError('join_type="cross" not supported. Please use "cartesian_product" task.')
    assert join_type in ['inner', 'outer', 'left', 'right']

    common_colnames = extract_common_colnames(table_1, table_2)

    if (on_cols is None and len(common_colnames) == 0) \
            or (on_cols is not None and len(on_cols) == 0):
        raise ValueError(f'No common column names were found. '
                         f'table_1: {tuple(table_1.columns)}. '
                         f'table_2: {tuple(table_2.columns)}. '
                         f'on_cols: {on_cols}')

    on = None
    left_on = None
    right_on = None

    if on_cols is None:
        on = common_colnames
    elif isinstance(on_cols, Mapping):
        left_on = tuple(on_cols.keys())
        right_on = tuple(on_cols.values())
    else:
        on = on_cols

    column_info = f'common columns: {on}' if on is not None \
        else f'column mappings: {tuple(on_cols.items())}'
    print(f'Joining tables on {column_info}, using join type: {join_type}...')

    merged_df = pd.merge(
        table_1.loc[:, :],
        table_2.loc[:, :],
        on=on,
        left_on=left_on,
        right_on=right_on,
        how=join_type,
        suffixes=('_1', '_2'),
    ).convert_dtypes()

    return PandasModel(merged_df)


@mypy_fix_task_template
@TaskTemplate()
def cartesian_product(table_1: PandasModel, table_2: PandasModel) -> PandasModel:
    merged_df = pd.merge(
        table_1.loc[:, :],
        table_2.loc[:, :],
        how='cross',
        suffixes=('_1', '_2'),
    ).convert_dtypes()

    return PandasModel(merged_df)



from typing import cast

from omnipy.api.protocols.public.compute import (IsDagFlowTemplate,
                                                 IsFuncFlowTemplate,
                                                 IsLinearFlowTemplate,
                                                 IsTaskTemplate)

# Needed due to lack of support for class decorators in mypy.
# See: https://github.com/python/mypy/issues/3135


def mypy_fix_task_template(task_template: object) -> 'IsTaskTemplate':
    return cast('IsTaskTemplate', task_template)


def mypy_fix_linear_flow_template(linear_flow_template: object) -> 'IsLinearFlowTemplate':
    return cast('IsLinearFlowTemplate', linear_flow_template)


def mypy_fix_dag_flow_template(dag_flow_template: object) -> 'IsDagFlowTemplate':
    return cast('IsDagFlowTemplate', dag_flow_template)


def mypy_fix_func_flow_template(func_flow_template: object) -> 'IsFuncFlowTemplate':
    return cast('IsFuncFlowTemplate', func_flow_template)



from typing import Generic

from typing_extensions import TypeVar

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

from .models import (JsonDictModel,
                     JsonDictOfDictsModel,
                     JsonDictOfDictsOfScalarsModel,
                     JsonDictOfListsModel,
                     JsonDictOfListsOfDictsModel,
                     JsonDictOfListsOfScalarsModel,
                     JsonDictOfNestedListsModel,
                     JsonDictOfScalarsModel,
                     JsonListModel,
                     JsonListOfDictsModel,
                     JsonListOfDictsOfScalarsModel,
                     JsonListOfListsModel,
                     JsonListOfListsOfScalarsModel,
                     JsonListOfNestedDictsModel,
                     JsonListOfScalarsModel,
                     JsonModel,
                     JsonNestedDictsModel,
                     JsonNestedListsModel,
                     JsonOnlyDictsModel,
                     JsonOnlyListsModel,
                     JsonScalarModel)

# TODO: switch from plural to singular for names of modules in omnipy modules
# TODO: call omnipy modules something else than modules, to distinguish from Python modules.
#       Perhaps plugins?
#
_JsonModelT = TypeVar('_JsonModelT', bound=Model, default=JsonModel)


class _JsonBaseDataset(Dataset[_JsonModelT], Generic[_JsonModelT]):
    """"""
    ...


class JsonDataset(_JsonBaseDataset[JsonModel]):
    """"""
    ...


class JsonScalarDataset(_JsonBaseDataset[JsonScalarModel]):
    """"""
    ...


# List at the top level


class JsonListDataset(_JsonBaseDataset[JsonListModel]):
    """"""
    ...


class JsonListOfScalarsDataset(_JsonBaseDataset[JsonListOfScalarsModel]):
    """"""
    ...


class JsonListOfListsDataset(_JsonBaseDataset[JsonListOfListsModel]):
    """"""
    ...


class JsonListOfListsOfScalarsDataset(_JsonBaseDataset[JsonListOfListsOfScalarsModel]):
    """"""
    ...


class JsonListOfDictsDataset(_JsonBaseDataset[JsonListOfDictsModel]):
    """"""
    ...


class JsonListOfDictsOfScalarsDataset(_JsonBaseDataset[JsonListOfDictsOfScalarsModel]):
    """"""
    ...


# Dict at the top level


class JsonDictDataset(_JsonBaseDataset[JsonDictModel]):
    ...


class JsonDictOfScalarsDataset(_JsonBaseDataset[JsonDictOfScalarsModel]):
    ...


class JsonDictOfListsDataset(_JsonBaseDataset[JsonDictOfListsModel]):
    ...


class JsonDictOfListsOfScalarsDataset(_JsonBaseDataset[JsonDictOfListsOfScalarsModel]):
    ...


class JsonDictOfDictsDataset(_JsonBaseDataset[JsonDictOfDictsModel]):
    ...


class JsonDictOfDictsOfScalarsDataset(_JsonBaseDataset[JsonDictOfDictsOfScalarsModel]):
    ...


# Nested datasets


class JsonOnlyListsDataset(_JsonBaseDataset[JsonOnlyListsModel]):
    ...


class JsonNestedListsDataset(_JsonBaseDataset[JsonNestedListsModel]):
    ...


class JsonOnlyDictsDataset(_JsonBaseDataset[JsonOnlyDictsModel]):
    ...


class JsonNestedDictsDataset(_JsonBaseDataset[JsonNestedDictsModel]):
    ...


# More specific datasets


class JsonListOfNestedDictsDataset(_JsonBaseDataset[JsonListOfNestedDictsModel]):
    ...


class JsonDictOfNestedListsDataset(_JsonBaseDataset[JsonDictOfNestedListsModel]):
    ...


class JsonDictOfListsOfDictsDataset(_JsonBaseDataset[JsonDictOfListsOfDictsModel]):
    ...


# Custom datasets



from abc import ABCMeta
from contextlib import AbstractContextManager
from datetime import datetime

from omnipy.api.protocols.private.compute.job_creator import IsJobCreator
from omnipy.api.protocols.private.engine import IsEngine
from omnipy.api.protocols.public.config import IsJobConfig


class JobCreator(AbstractContextManager):
    def __init__(self) -> None:
        self._engine: IsEngine | None = None
        self._config: IsJobConfig | None = None
        self._nested_context_level: int = 0
        self._time_of_cur_toplevel_nested_context_run: datetime | None = None

    def set_engine(self, engine: IsEngine) -> None:
        self._engine = engine

    def set_config(self, config: IsJobConfig) -> None:
        self._config = config

    def __enter__(self):
        if self._nested_context_level == 0:
            self._time_of_cur_toplevel_nested_context_run = datetime.now()

        self._nested_context_level += 1

    def __exit__(self, exc_type, exc_value, traceback):
        self._nested_context_level -= 1

        if self._nested_context_level == 0:
            self._time_of_cur_toplevel_nested_context_run = None

    @property
    def engine(self) -> IsEngine | None:
        return self._engine

    @property
    def config(self) -> IsJobConfig | None:
        return self._config

    @property
    def nested_context_level(self) -> int:
        return self._nested_context_level

    @property
    def time_of_cur_toplevel_nested_context_run(self) -> datetime | None:
        return self._time_of_cur_toplevel_nested_context_run


class JobBaseMeta(ABCMeta):
    """"""
    _job_creator_obj = JobCreator()

    @property
    def job_creator(self) -> IsJobCreator:
        return self._job_creator_obj

    @property
    def nested_context_level(self) -> int:
        return self.job_creator.nested_context_level



import asyncio
from inspect import iscoroutinefunction
from typing import Callable, cast, Coroutine

from omnipy.api.protocols.private.compute.job import IsJobBase, IsPlainFuncArgJobBase
from omnipy.util.helpers import get_event_loop_and_check_if_loop_is_running


class AutoAsyncJobBaseMixin:
    def __init__(
        self,
        *,
        auto_async: bool = True,
    ):
        self._auto_async = auto_async

    @property
    def auto_async(self) -> bool:
        return self._auto_async

    def _call_job(self, *args: object, **kwargs: object) -> object:
        self_as_plain_func_arg_job_base = cast(IsPlainFuncArgJobBase, self)
        self_as_job_base = cast(IsJobBase, self)
        super_as_job_base = cast(IsJobBase, super())

        job_func = self_as_plain_func_arg_job_base._job_func
        job_func_is_coroutine: bool = iscoroutinefunction(job_func)
        super_call_job_func = cast(Callable[..., Coroutine], super_as_job_base._call_job)

        loop, loop_is_running = get_event_loop_and_check_if_loop_is_running()
        if self.auto_async and job_func_is_coroutine and not self_as_job_base.in_flow_context:
            if loop and loop_is_running:
                res: object = loop.create_task(super_call_job_func(*args, **kwargs))
            else:
                res = asyncio.run(super_call_job_func(*args, **kwargs))
        else:
            res = super_call_job_func(*args, **kwargs)

        return res



import asyncio
from collections import defaultdict, UserDict
from collections.abc import Iterable, Mapping, MutableMapping
from copy import copy
import json
import os
import tarfile
from typing import Any, Callable, cast, Generic, Iterator, TYPE_CHECKING

from pydantic import Field, PrivateAttr, root_validator, ValidationError
from pydantic.fields import ModelField, Undefined, UndefinedType
from pydantic.generics import GenericModel
from pydantic.main import ModelMetaclass
from typing_extensions import TypeVar

from omnipy.api.typedefs import TypeForm
from omnipy.data.data_class_creator import DataClassBase, DataClassBaseMeta
from omnipy.data.helpers import (cleanup_name_qualname_and_module,
                                 is_model_instance,
                                 is_model_subclass)
from omnipy.data.mixins.display import DatasetDisplayMixin
from omnipy.data.mixins.task import TaskDatasetMixin
from omnipy.data.model import Model
from omnipy.data.selector import (create_updated_mapping,
                                  Index2DataItemsType,
                                  Key2DataItemType,
                                  prepare_selected_items_with_iterable_data,
                                  prepare_selected_items_with_mapping_data,
                                  select_keys)
from omnipy.util.decorators import call_super_if_available
from omnipy.util.helpers import (get_default_if_typevar,
                                 get_event_loop_and_check_if_loop_is_running,
                                 is_iterable,
                                 remove_forward_ref_notation)
from omnipy.util.web import download_file_to_memory

if TYPE_CHECKING:
    from omnipy.modules.remote.datasets import HttpUrlDataset
    from omnipy.modules.remote.models import HttpUrlModel

ModelT = TypeVar('ModelT', bound=Model)
GeneralModelT = TypeVar('GeneralModelT', bound=Model)
_DatasetT = TypeVar('_DatasetT')

DATA_KEY = 'data'
ASYNC_LOAD_SLEEP_TIME = 0.05

# def orjson_dumps(v, *, default):
#     # orjson.dumps returns bytes, to match standard json.dumps we need to decode
#     return orjson.dumps(v, default=default).decode()

# TODO: implement copy(), __copy__() and __deepcopy__() for Dataset and Model, making use of
#       BaseModel.copy()


class _DatasetMetaclass(DataClassBaseMeta, ModelMetaclass):
    ...


class Dataset(
        DatasetDisplayMixin,
        TaskDatasetMixin,
        DataClassBase,
        GenericModel,
        Generic[ModelT],
        UserDict,
        metaclass=_DatasetMetaclass):
    """
    Dict-based container of data files that follow a specific Model

    Dataset is a generic class that cannot be instantiated directly. Instead, a Dataset class needs
    to be specialized with a data model before Dataset objects can be instantiated. A data model
    functions as a data parser and guarantees that the parsed data follows the specified model.

    The specialization must be done through the use of Model, either directly, e.g.::

        MyDataset = Dataset[Model[dict[str, list[int]]])

    ... or indirectly, using a Model subclass, e.g.::

        class MyModel(Model[dict[str, list[int]]):
            pass

        MyDataset = Dataset[MyModel]

    ... alternatively through the specification of a Dataset subclass::

        class MyDataset(Dataset[MyModel]):
            pass

    The specialization can also be done in a more deeply nested structure, e.g.::

        class MyNumberList(Model[list[int]]):
            pass

        class MyToplevelDict(Model[dict[str, MyNumberList]]):
            pass

        class MyDataset(Dataset[MyToplevelDict]):
            pass

    Once instantiated, a dataset object functions as a dict of data files, with the keys
    referring to the data file names and the contents to the data file contents, e.g.::

        MyNumberListDataset = Dataset[Model[list[int]]]

        my_dataset = MyNumberListDataset({'file_1': [1,2,3]})
        my_dataset['file_2'] = [2,3,4]

        print(my_dataset.keys())

    The Dataset class is a wrapper class around the powerful `GenericModel` class from pydantic.
    """
    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

        # TODO: Use json serializer package from the pydantic config instead of 'json'

        # json_loads = orjson.loads
        # json_dumps = orjson_dumps

    data: dict[str, ModelT] = Field(default={})

    def __class_getitem__(
        cls,
        params: type[ModelT] | tuple[type[ModelT]] | tuple[type[ModelT], Any] | TypeVar
        | tuple[TypeVar, ...],
    ) -> 'type[Dataset[type[ModelT]]]':
        # TODO: change model type to params: Type[Any] | tuple[Type[Any], ...]
        #       as in GenericModel.

        model = cls._prepare_params(params)
        orig_model = cls._clean_model(model)

        if cls == Dataset:
            if not isinstance(orig_model, TypeVar) and not is_model_subclass(orig_model):
                raise TypeError('Invalid model: {}! '.format(orig_model)
                                + 'omnipy Dataset models must be a specialization of the omnipy '
                                'Model class.')
        else:
            if isinstance(orig_model, TypeVar):
                model = get_default_if_typevar(orig_model)

        created_dataset = super().__class_getitem__(model)
        cls._recursively_set_allow_none(created_dataset._get_data_field())
        cleanup_name_qualname_and_module(cls, created_dataset, orig_model)

        return created_dataset

    @call_super_if_available(call_super_before_method=True)
    @classmethod
    def _clean_model(cls, model: TypeForm) -> TypeForm:
        return model

    def __init__(  # noqa: C901
        self,
        value: Mapping[str, object] | Iterable[tuple[str, object]] | UndefinedType = Undefined,
        *,
        data: Mapping[str, object] | UndefinedType = Undefined,
        **kwargs: object,
    ) -> None:
        # TODO: Error message when forgetting parenthesis when creating Dataset should be improved.
        #       Unclear where this can be done, if anywhere? E.g.:
        #           a = Dataset[Model[int]]
        #           a['adsfas'] = 2
        #           Traceback (most recent call last):
        #             ...
        #           TypeError: 'ModelMetaclass' object does not support item assignment
        #
        # TODO: Disallow e.g.:
        #       Dataset[Model[str]](Model[int](5)) ==  Dataset[Model[str]](data=Model[int](5))
        #       == Dataset[Model[str]](data={'__root__': Model[str]('5')})

        super_kwargs = {}

        assert DATA_KEY not in kwargs, \
            ('Not allowed with "data" as kwargs key. Not sure how you managed this? Are you trying '
             'to break Dataset init on purpose?')

        if value != Undefined:
            assert data == Undefined, \
                'Not allowed to combine positional and "data" keyword argument'
            assert len(kwargs) == 0, \
                'Not allowed to combine positional and keyword arguments'
            super_kwargs[DATA_KEY] = value

        if data != Undefined:
            assert len(kwargs) == 0, \
                f"Not allowed to combine '{DATA_KEY}' with other keyword arguments"
            super_kwargs[DATA_KEY] = data

        if kwargs:
            if DATA_KEY not in super_kwargs:
                super_kwargs[DATA_KEY] = kwargs
                kwargs = {}

        model_cls = self.get_model_class()
        if model_cls == ModelT:  # type: ignore[misc]
            self._raise_no_model_exception()

        model_or_dataset_as_input = False
        if DATA_KEY in super_kwargs:
            done = False
            while not done:
                match super_kwargs[DATA_KEY]:
                    case Dataset():
                        model_or_dataset_as_input = True
                        super_kwargs[DATA_KEY] = cast(Dataset, super_kwargs[DATA_KEY]).to_data()
                        done = True
                    case MutableMapping():
                        mapping_value = cast(MutableMapping, super_kwargs[DATA_KEY])
                        for key, val in mapping_value.items():
                            if is_model_instance(val):
                                model_or_dataset_as_input = True
                                mapping_value[key] = model_cls(val)
                        done = True
                    case Iterable():
                        try:
                            super_kwargs[DATA_KEY] = dict(cast(Iterable, super_kwargs[DATA_KEY]))
                        except TypeError as e:
                            raise TypeError('Data object must be a mapping or an iterable of '
                                            '(key, val) pairs') from e
                    case _:
                        done = True

        self._init(super_kwargs, **kwargs)

        try:
            self._primary_validation(super_kwargs)
        except ValidationError:
            if model_or_dataset_as_input:
                self._secondary_validation_from_data(super_kwargs)
            else:
                raise

        if not self.__doc__:
            self._set_standard_field_description()

    def _primary_validation(self, super_kwargs):
        # Pydantic validation of super_kwargs
        super().__init__(**super_kwargs)

    def _secondary_validation_from_data(self, super_kwargs):
        super().__init__()
        self.from_data(super_kwargs[DATA_KEY])

    def _init(self, super_kwargs: dict[str, Any], **kwargs: Any) -> None:
        ...

    # TODO: Revise with pydantic v2: __deepcopy__ is not defined for Dataset and Model, as it is not
    #       supported by pydantic v1. BaseModel.copy(deep=True) does not support a deepcopy memo.
    #       So we instead make use of the builtin support for deepcopy, which seems to work fine.
    #       However, __deepcopy__ in pydantic v2 is probably more efficient due to the memo and
    #       the Rust backend.

    def __copy__(self):
        return self.copy(deep=False)

    def copy(self, *, deep: bool = False, **kwargs) -> 'Dataset[ModelT]':
        pydantic_copy = GenericModel.copy(self, deep=deep, **kwargs)
        if not deep:
            pydantic_copy.__dict__[DATA_KEY] = pydantic_copy.__dict__[DATA_KEY].copy()
        return pydantic_copy

    @classmethod
    def clone_dataset_cls(cls: type[_DatasetT],
                          new_dataset_cls_name: str,
                          model_cls: type | None = None) -> type[_DatasetT]:
        # TODO: Update `model_cls` type to `type[IsModel] | None` when IsModel is defined
        if model_cls:
            generic_dataset_cls = cls.__bases__[0]
            new_base_cls = generic_dataset_cls[model_cls]
        else:
            new_base_cls = cls

        new_dataset_cls: type[_DatasetT] = type(new_dataset_cls_name, (new_base_cls,), {})
        return new_dataset_cls

    @classmethod
    def _get_data_field(cls) -> ModelField:
        return cast(ModelField, cls.__fields__.get(DATA_KEY))

    @classmethod
    def get_model_class(cls) -> type[Model]:
        """
        Returns the concrete Model class used for all data files in the dataset, e.g.:
        `Model[list[int]]`
        :return: The concrete Model class used for all data files in the dataset
        """
        return cls._clean_model(cls._get_data_field().type_)

    @staticmethod
    def _raise_no_model_exception() -> None:
        raise TypeError(
            'Note: The Dataset class requires a Model class (or a subclass) to be specified as '
            'a type hierarchy within brackets either directly, e.g.:\n\n'
            '\tmodel = Dataset[Model[list[int]]]()\n\n'
            'or indirectly in a subclass definition, e.g.:\n\n'
            '\tclass MyNumberListDataset(Dataset[Model[list[int]]]): ...\n\n'
            'For anything other than the simplest cases, the definition of Model and Dataset '
            'subclasses is encouraged , e.g.:\n\n'
            '\tclass MyNumberListModel(Model[list[int]]): ...\n'
            '\tclass MyDataset(Dataset[MyNumberListModel]): ...\n\n')

    def _set_standard_field_description(self) -> None:
        self.__fields__[DATA_KEY].field_info.description = self._get_standard_field_description()

    @classmethod
    def _get_standard_field_description(cls) -> str:
        return ('This class represents a dataset in the `omnipy` Python package and contains '
                'a set of named data items that follows the same data model. '
                'It is a statically typed specialization of the Dataset class according to a '
                'particular specialization of the Model class. Both main classes are wrapping '
                'the excellent Python package named `pydantic`.')

    def __getitem__(self, selector: str | int | slice | Iterable[str | int]) -> Any:
        selected_keys = select_keys(selector, self.data)

        if selected_keys.singular:
            value = self.data[selected_keys.keys[0]]
        else:
            value = self.__class__({key: self.data[key] for key in selected_keys.keys})

        return self._check_value(value)

    @call_super_if_available(call_super_before_method=True)
    def _check_value(self, value: Any) -> Any:
        return value

    def __delitem__(self, selector: str | int | slice | Iterable[str | int]) -> Any:
        selected_keys = select_keys(selector, self.data)

        if selected_keys.singular:
            del self.data[selected_keys.keys[0]]
        else:
            prev_data = copy(self.data)

            try:
                for key in selected_keys.keys:
                    del self.data[key]
            except Exception:
                self.data = prev_data
                raise

    def __setitem__(
        self,
        selector: str | int | slice | Iterable[str | int],
        data_obj: dict[str, ModelT] | Iterable[ModelT] | ModelT,
    ) -> None:
        selected_keys = select_keys(selector, self.data)

        if selected_keys.singular:
            self._set_data_file_and_validate(selected_keys.keys[0], cast(ModelT, data_obj))
        else:
            key_2_data_item: Key2DataItemType[ModelT]
            index_2_data_items: Index2DataItemsType[ModelT]

            if isinstance(data_obj, MutableMapping):
                key_2_data_item, index_2_data_items = \
                    prepare_selected_items_with_mapping_data(
                        selected_keys.keys, selected_keys.last_index, data_obj,)

            elif is_iterable(data_obj) and not isinstance(data_obj, (str, bytes)):
                key_2_data_item, index_2_data_items = \
                    prepare_selected_items_with_iterable_data(
                        selected_keys.keys, selected_keys.last_index, tuple(data_obj), self.data)

            else:
                raise TypeError('Data object must be a mapping or an iterable')

            self._update_selected_items_with_data_items(key_2_data_item, index_2_data_items)

    def _update_selected_items_with_data_items(
        self,
        key_2_data_item: Key2DataItemType[ModelT],
        index_2_data_item: Index2DataItemsType[ModelT],
    ) -> None:

        updated_mapping = create_updated_mapping(self.data, key_2_data_item, index_2_data_item)
        self._replace_data_with_mapping(updated_mapping)

    def _replace_data_with_mapping(self, updated_mapping):
        prev_data = self.data
        try:
            self.absorb_and_replace(self.__class__(updated_mapping))
        except Exception:
            self.data = prev_data
            raise

    def _set_data_file_and_validate(self, key: str, val: ModelT) -> None:
        has_prev_value = key in self.data
        if has_prev_value:
            prev_value = self.data[key]

        try:
            self.data[key] = val
            self._validate_data_file(key)
        except Exception:
            if has_prev_value:
                self.data[key] = prev_value
            else:
                del self.data[key]
            raise

    @classmethod
    def update_forward_refs(cls, **localns: Any) -> None:
        """
        Try to update ForwardRefs on fields based on this Model, globalns and localns.
        """
        cls.get_model_class().update_forward_refs(**localns)  # Update Model cls
        super().update_forward_refs(**localns)
        cls.__name__ = remove_forward_ref_notation(cls.__name__)
        cls.__qualname__ = remove_forward_ref_notation(cls.__qualname__)

    def _validate_data_file(self, _data_file: str) -> None:
        val = self.data[_data_file]
        if is_model_instance(val):
            self.data[_data_file] = self.get_model_class()(val)
        else:
            self._force_full_validation()

    def _force_full_validation(self):
        self.data = self.data  # Triggers pydantic validation, as validate_assignment=True

    def __iter__(self) -> Iterator:
        return UserDict.__iter__(self)

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in self.__dict__ or attr == DATA_KEY or attr.startswith('__'):
            super().__setattr__(attr, value)
        elif attr == 'repr_state':
            prop = getattr(self.__class__, attr)
            prop.__set__(self, value)
        else:
            raise RuntimeError('Model does not allow setting of extra attributes')

    @root_validator
    def _parse_root_object(cls, root_obj: dict[str, ModelT]) -> Any:  # noqa
        assert DATA_KEY in root_obj
        data_dict = root_obj[DATA_KEY]
        model = cls.get_model_class()
        for key, val in data_dict.items():
            if val is None:
                data_dict[key] = model.parse_obj(val)
        return {DATA_KEY: data_dict}

    def to_data(self) -> dict[str, Any]:
        return {
            key: self._check_value(val) for key,
            val in GenericModel.dict(self, by_alias=True).get(DATA_KEY).items()
        }

    def from_data(self,
                  data: dict[str, Any] | Iterator[tuple[str, Any]],
                  update: bool = True) -> None:
        def callback_func(model: ModelT, contents: Any):
            model.from_data(contents)

        self._from_dict_with_callback(data, update, callback_func)

    def _from_dict_with_callback(self,
                                 data: dict[str, Any] | Iterator[tuple[str, Any]],
                                 update: bool,
                                 callback_func: Callable):
        if not isinstance(data, dict):
            data = dict(data)

        if not update:
            self.clear()

        model_cls = self.get_model_class()
        for data_file, contents in data.items():
            new_model = model_cls()
            callback_func(new_model, contents)
            self.data[data_file] = new_model

    def absorb(self, other: 'Dataset'):
        self.from_data(other.to_data(), update=True)

    def absorb_and_replace(self, other: 'Dataset'):
        self.from_data(other.to_data(), update=False)

    def to_json(self, pretty=True) -> dict[str, str]:
        result = {}

        for key, val in self.to_data().items():
            result[key] = self._pretty_print_json(val) if pretty else json.dumps(val)
        return result

    def from_json(self,
                  data: Mapping[str, str] | Iterable[tuple[str, str]],
                  update: bool = True) -> None:
        def callback_func(model: ModelT, contents: Any):
            model.from_json(contents)

        self._from_dict_with_callback(data, update, callback_func)

    # @classmethod
    # def get_type_args(cls):
    #     return cls.__fields__.get(DATA_KEY).type_
    #
    #
    # @classmethod
    # def create_from_json(cls, data: str, tuple[str]]):
    #     if isinstance(data, tuple):
    #         data = data[0]
    #
    #     obj = cls()
    #     obj.from_json(data, update=False)
    #     return obj
    #
    # def __reduce__(self):
    #     return self.__class__.create_from_json, (self.to_json(),)

    @classmethod
    def to_json_schema(cls, pretty: bool = True) -> str | dict[str, str]:
        result = {}
        clean_dataset = super(Dataset, Dataset).__class_getitem__(cls.get_model_class())
        schema = clean_dataset.schema()
        for key, val in schema['properties'][DATA_KEY].items():
            result[key] = val
        result['title'] = cls.__name__
        result['definitions'] = schema['definitions']

        for model_desc in result['definitions'].values():
            if 'orig_model' in model_desc:
                del model_desc['orig_model']

        if pretty:
            return cls._pretty_print_json(result)
        else:
            return json.dumps(result)

    @staticmethod
    def _pretty_print_json(json_content: Any) -> str:
        return json.dumps(json_content, indent=2)

    def save(self, path: str):
        serializer_registry = self._get_serializer_registry()

        parsed_dataset, serializer = serializer_registry.auto_detect_tar_file_serializer(self)

        if serializer is None:
            print(f'Unable to find a serializer for dataset with data type "{type(self)}". '
                  f'Will abort saving...')
        else:
            if not path.endswith('.tar.gz'):
                out_tar_gz_path = f'{path}.tar.gz'

            print(f'Writing dataset as a gzipped tarpack to "{os.path.abspath(out_tar_gz_path)}"')

            with open(out_tar_gz_path, 'wb') as out_tar_gz_file:
                out_tar_gz_file.write(serializer.serialize(parsed_dataset))

            directory = os.path.abspath(out_tar_gz_path[:-7])
            if not os.path.exists(directory):
                os.makedirs(directory)

            tar = tarfile.open(out_tar_gz_path)
            print(f'Extracting contents to directory "{os.path.abspath(out_tar_gz_path[:-7])}"')
            tar.extractall(path=directory)
            tar.close()

    def load(self,
             paths_or_urls: 'str | Iterable[str] | HttpUrlModel | HttpUrlDataset',
             by_file_suffix: bool = False) -> list[asyncio.Task] | None:
        from omnipy import HttpUrlDataset, HttpUrlModel

        match paths_or_urls:
            case HttpUrlDataset():
                return self._load_http_urls(paths_or_urls)

            case HttpUrlModel():
                return self._load_http_urls(HttpUrlDataset({str(paths_or_urls): paths_or_urls}))

            case str():
                try:
                    http_url_dataset = HttpUrlDataset({paths_or_urls: paths_or_urls})
                except ValidationError:
                    return self._load_paths([paths_or_urls], by_file_suffix)
                return self._load_http_urls(http_url_dataset)
            case Iterable():
                try:
                    path_or_url_iterable = cast(Iterable[str], paths_or_urls)
                    http_url_dataset = HttpUrlDataset(
                        zip(path_or_url_iterable, path_or_url_iterable))
                except ValidationError:
                    return self._load_paths(path_or_url_iterable, by_file_suffix)
                return self._load_http_urls(http_url_dataset)
            case _:
                raise TypeError(f'"paths_or_urls" argument is of incorrect type. Type '
                                f'{type(paths_or_urls)} is not supported.')

    def _load_http_urls(self, http_url_dataset: 'HttpUrlDataset') -> list[asyncio.Task]:
        from omnipy.modules.remote.helpers import RateLimitingClientSession
        from omnipy.modules.remote.tasks import get_json_from_api_endpoint
        hosts: defaultdict[str, list[int]] = defaultdict(list)
        for i, url in enumerate(http_url_dataset.values()):
            hosts[url.host].append(i)

        async def load_all():
            tasks = []

            for host in hosts:
                async with RateLimitingClientSession(
                        self.config.http_config_for_host[host].requests_per_time_period,
                        self.config.http_config_for_host[host].time_period_in_secs
                ) as client_session:
                    indices = hosts[host]
                    task = (
                        get_json_from_api_endpoint.refine(
                            output_dataset_param='output_dataset').run(
                                http_url_dataset[indices],
                                client_session=client_session,
                                output_dataset=self))
                    tasks.append(task)
                    while not task.done():
                        await asyncio.sleep(ASYNC_LOAD_SLEEP_TIME)

            await asyncio.gather(*tasks)
            return self

        loop, loop_is_running = get_event_loop_and_check_if_loop_is_running()

        if loop and loop_is_running:
            return loop.create_task(load_all())
        else:
            return asyncio.run(load_all())

    def _load_paths(self, path_or_urls: Iterable[str], by_file_suffix: bool) -> None:
        for path_or_url in path_or_urls:
            serializer_registry = self._get_serializer_registry()
            tar_gz_file_path = self._ensure_tar_gz_file(path_or_url)

            if by_file_suffix:
                loaded_dataset = \
                    serializer_registry.load_from_tar_file_path_based_on_file_suffix(
                        self, tar_gz_file_path, self)
            else:
                loaded_dataset = \
                    serializer_registry.load_from_tar_file_path_based_on_dataset_cls(
                        self, tar_gz_file_path, self)
            if loaded_dataset is not None:
                self.absorb(loaded_dataset)
                continue
            else:
                raise RuntimeError('Unable to load from serializer')

    @staticmethod
    def _download_file(url: str, path: str, tmp_dir_path: str) -> str | None:
        print(f'Downloading {url}...')
        data = download_file_to_memory(url)

        if data is None:
            print(f'Failed to download file from {url}')
            return None

        download_path = os.path.join(tmp_dir_path, os.path.basename(path))
        with open(download_path, 'wb') as out_file:
            out_file.write(data)
        return download_path

    @staticmethod
    def _ensure_tar_gz_file(path: str):
        assert os.path.exists(path), f'No file or directory at {path}'

        if not path.endswith('.tar.gz'):
            tar_gz_file_path = path + '.tar.gz'
            if not os.path.isfile(tar_gz_file_path):
                print(f'Creating compressed file {os.path.abspath(tar_gz_file_path)} from '
                      f'the contents of "{os.path.abspath(path)}"')

                with tarfile.open(tar_gz_file_path, 'w:gz') as tar:
                    if os.path.isdir(path):
                        for fn in sorted(os.listdir(path)):
                            p = os.path.join(path, fn)
                            tar.add(p, arcname=fn)
                    elif os.path.isfile(path):
                        tar.add(path, arcname=os.path.basename(path))
            return tar_gz_file_path

        return path

    @staticmethod
    def _get_serializer_registry():
        from omnipy.modules import get_serializer_registry
        return get_serializer_registry()

    def as_multi_model_dataset(self) -> 'MultiModelDataset[ModelT]':
        multi_model_dataset = MultiModelDataset[self.get_model_class()]()
        for data_file in self:
            multi_model_dataset.data[data_file] = self.data[data_file]
        return multi_model_dataset

    def __eq__(self, other: object) -> bool:
        # return self.__class__ == other.__class__ and super().__eq__(other)
        return isinstance(other, Dataset) \
            and self.__class__ == other.__class__ \
            and self.data == other.data \
            and self.to_data() == other.to_data()  # last is probably unnecessary, but just in case

    def __repr_args__(self):
        return [(k, v.contents) if is_model_instance(v) else (k, v) for k, v in self.data.items()]


class MultiModelDataset(Dataset[GeneralModelT], Generic[GeneralModelT]):
    """
        Variant of Dataset that allows custom models to be set on individual data files

        Note that the general model still needs to hold for all data files, in addition to any
        custom models.
    """

    _custom_field_models: dict[str, GeneralModelT] = PrivateAttr(default={})

    def set_model(self, data_file: str, model: GeneralModelT) -> None:
        try:
            self._custom_field_models[data_file] = model
            if data_file in self.data:
                self._validate_data_file(data_file)
            else:
                self.data[data_file] = model()
        except ValidationError:
            del self._custom_field_models[data_file]
            raise

    def get_model(self, data_file: str) -> GeneralModelT:
        if data_file in self._custom_field_models:
            return self._custom_field_models[data_file]
        else:
            return self.get_model_class()

    def from_data(self,
                  data: dict[str, Any] | Iterator[tuple[str, Any]],
                  update: bool = True) -> None:
        super().from_data(data, update)
        for data_file in self:
            self._validate_data_file_according_to_custom_field_model(data_file)
        self._force_full_validation()

    def _validate_data_file(self, data_file: str) -> None:
        self._validate_data_file_according_to_custom_field_model(data_file)
        self._force_full_validation()

    def _validate_data_file_according_to_custom_field_model(self, data_file: str):
        if data_file in self._custom_field_models:
            model = self._custom_field_models[data_file]
            if not is_model_instance(model):
                model = Model[model]
            data_obj = self._to_data_if_model(self.data[data_file])
            parsed_data = self._to_data_if_model(model(data_obj))
            self.data[data_file] = parsed_data

    @staticmethod
    def _to_data_if_model(data_obj: Any):
        if is_model_instance(data_obj):
            data_obj = data_obj.to_data()
        return data_obj



from abc import ABCMeta
from contextlib import contextmanager
from typing import Callable, ContextManager, Iterator

from pydantic.fields import ModelField

from omnipy.api.protocols.private.data import IsDataClassCreator
from omnipy.api.protocols.private.util import HasContents, IsSnapshotHolder
from omnipy.api.protocols.public.config import IsDataConfig
from omnipy.api.typedefs import TypeForm
from omnipy.config.data import DataConfig
from omnipy.util.decorators import call_super_if_available
from omnipy.util.helpers import is_union, SnapshotHolder


class DataClassCreator:
    def __init__(self) -> None:
        self._config: IsDataConfig = DataConfig()
        self._snapshot_holder = SnapshotHolder[HasContents, object]()
        self._deepcopy_context_level = 0

    @property
    def config(self) -> IsDataConfig:
        return self._config

    def set_config(self, config: IsDataConfig) -> None:
        self._config = config

    @property
    def snapshot_holder(self) -> IsSnapshotHolder[HasContents, object]:
        return self._snapshot_holder

    def deepcopy_context(
        self,
        top_level_entry_func: Callable[[], None],
        top_level_exit_func: Callable[[], None],
    ) -> ContextManager[int]:
        @contextmanager
        def _call_exit_func_if_top_level(*args, **kwds) -> Iterator[int]:
            if self._deepcopy_context_level == 0:
                top_level_entry_func()

            self._deepcopy_context_level += 1

            try:
                yield self._deepcopy_context_level
            finally:
                self._deepcopy_context_level -= 1

                if self._deepcopy_context_level == 0:
                    top_level_exit_func()

        return _call_exit_func_if_top_level()


class DataClassBaseMeta(ABCMeta):
    """"""
    _data_class_creator_obj: IsDataClassCreator = DataClassCreator()

    @property
    def data_class_creator(self) -> IsDataClassCreator:
        return self._data_class_creator_obj


class DataClassBase(metaclass=DataClassBaseMeta):
    @call_super_if_available(call_super_before_method=False)
    @classmethod
    def _prepare_params(cls, params: TypeForm) -> TypeForm:
        # This line is needed for interoperability with pydantic GenericModel, which internally
        # stores the model as a len(1) tuple
        return params[0] if isinstance(params, tuple) and len(params) == 1 else params

    @classmethod
    def _recursively_set_allow_none(cls, field: ModelField) -> None:
        if field.sub_fields:
            if is_union(field.outer_type_):
                if any(_.allow_none for _ in field.sub_fields):
                    field.allow_none = True

            for sub_field in field.sub_fields:
                cls._recursively_set_allow_none(sub_field)
        if field.key_field:
            if is_union(field.key_field.outer_type_):
                if any(_.allow_none for _ in field.key_field.sub_fields):
                    field.key_field.allow_none = True

            if field.key_field.sub_fields:
                for sub_field in field.key_field.sub_fields:
                    cls._recursively_set_allow_none(sub_field)

    @property
    def _data_class_creator(self) -> IsDataClassCreator:
        return self.__class__.data_class_creator

    @property
    def config(self) -> IsDataConfig:
        return self.__class__.data_class_creator.config

    @property
    def snapshot_holder(self) -> IsSnapshotHolder[HasContents, object]:
        return self.__class__.data_class_creator.snapshot_holder

    def deepcopy_context(
        self,
        top_level_entry_func: Callable[[], None],
        top_level_exit_func: Callable[[], None],
    ) -> ContextManager[int]:
        return self.__class__.data_class_creator.deepcopy_context(top_level_entry_func,
                                                                  top_level_exit_func)



from dataclasses import dataclass
from datetime import timedelta
from typing import Any, Callable, Type

from omnipy.api.protocols.public.compute import IsDagFlow, IsFuncFlow, IsLinearFlow, IsTask
from omnipy.api.protocols.public.config import IsPrefectEngineConfig
from omnipy.config.engine import PrefectEngineConfig
from omnipy.engine.job_runner import (DagFlowRunnerEngine,
                                      FuncFlowRunnerEngine,
                                      LinearFlowRunnerEngine,
                                      TaskRunnerEngine)
from omnipy.util.helpers import resolve
from omnipy.util.publisher import RuntimeEntryPublisher

from .. import prefect_flow, prefect_task, PrefectFlow, PrefectTask, task_input_hash


@dataclass
class PrefectEngineConfigEntryPublisher(PrefectEngineConfig, RuntimeEntryPublisher):
    ...


class PrefectEngine(TaskRunnerEngine,
                    LinearFlowRunnerEngine,
                    DagFlowRunnerEngine,
                    FuncFlowRunnerEngine):
    """Job runner engine for Prefect"""
    def _init_engine(self) -> None:
        ...

    def _update_from_config(self) -> None:
        ...

    @classmethod
    def get_config_cls(cls) -> Type[IsPrefectEngineConfig]:
        return PrefectEngineConfig

    # TaskRunnerEngine

    def _init_task(self, task: IsTask, call_func: Callable) -> PrefectTask:
        assert isinstance(self._config, PrefectEngineConfig)
        task_kwargs = dict(
            name=task.name,
            cache_key_fn=task_input_hash if self._config.use_cached_results else None,
            cache_expiration=timedelta(days=1))

        if task.has_coroutine_func():

            @prefect_task(**task_kwargs)
            async def run_task(*inner_args, **inner_kwargs):
                return await call_func(*inner_args, **inner_kwargs)
        else:

            @prefect_task(**task_kwargs)
            def run_task(*inner_args, **inner_kwargs):
                return call_func(*inner_args, **inner_kwargs)

        return run_task

    def _run_task(self, state: PrefectTask, task: IsTask, call_func: Callable, *args,
                  **kwargs) -> Any:

        _prefect_task = state

        if task.in_flow_context:
            return _prefect_task(*args, **kwargs)
        else:
            flow_kwargs = dict(name=task.name)

            if task.has_coroutine_func():

                @prefect_flow(**flow_kwargs)
                async def task_flow(*inner_args, **inner_kwargs):
                    return await resolve(_prefect_task(*inner_args, **inner_kwargs))
            else:

                @prefect_flow(**flow_kwargs)
                def task_flow(*inner_args, **inner_kwargs):
                    return _prefect_task(*inner_args, **inner_kwargs)

            return task_flow(*args, **kwargs)

    # LinearFlowRunnerEngine
    def _init_linear_flow(self, linear_flow: IsLinearFlow) -> Any:
        assert isinstance(self._config, PrefectEngineConfig)
        flow_kwargs = dict(name=linear_flow.name,)
        call_func = self.default_linear_flow_run_decorator(linear_flow)

        if linear_flow.has_coroutine_func():

            @prefect_flow(**flow_kwargs)
            async def run_linear_flow(*inner_args, **inner_kwargs):
                with linear_flow.flow_context:
                    return await resolve(call_func(*inner_args, **inner_kwargs))
        else:

            @prefect_flow(**flow_kwargs)
            def run_linear_flow(*inner_args, **inner_kwargs):
                with linear_flow.flow_context:
                    return call_func(*inner_args, **inner_kwargs)

        return run_linear_flow

    def _run_linear_flow(self, state: Any, linear_flow: IsLinearFlow, *args, **kwargs) -> Any:

        _prefect_flow = state

        return _prefect_flow(*args, **kwargs)

    # DagFlowRunnerEngine

    def _init_dag_flow(self, dag_flow: IsDagFlow) -> Any:

        assert isinstance(self._config, PrefectEngineConfig)
        flow_kwargs = dict(name=dag_flow.name,)
        call_func = self.default_dag_flow_run_decorator(dag_flow)

        if dag_flow.has_coroutine_func():

            @prefect_flow(**flow_kwargs)
            async def run_dag_flow(*inner_args, **inner_kwargs):
                with dag_flow.flow_context:
                    return await resolve(call_func(*inner_args, **inner_kwargs))
        else:

            @prefect_flow(**flow_kwargs)
            def run_dag_flow(*inner_args, **inner_kwargs):
                with dag_flow.flow_context:
                    return call_func(*inner_args, **inner_kwargs)

        return run_dag_flow

    def _run_dag_flow(self, state: PrefectFlow, flow: IsDagFlow, *args, **kwargs) -> Any:

        _prefect_flow = state

        return _prefect_flow(*args, **kwargs)

    # FuncFlowRunnerEngine

    def _init_func_flow(self, func_flow: IsFuncFlow, call_func: Callable) -> object:

        assert isinstance(self._config, PrefectEngineConfig)
        flow_kwargs = dict(name=func_flow.name,)

        if func_flow.has_coroutine_func():

            @prefect_flow(**flow_kwargs)
            async def run_func_flow(*inner_args, **inner_kwargs):
                with func_flow.flow_context:
                    return await resolve(call_func(*inner_args, **inner_kwargs))
        else:

            @prefect_flow(**flow_kwargs)
            def run_func_flow(*inner_args, **inner_kwargs):
                with func_flow.flow_context:
                    return call_func(*inner_args, **inner_kwargs)

        return run_func_flow

    def _run_func_flow(self,
                       state: Any,
                       func_flow: IsFuncFlow,
                       call_func: Callable,
                       *args,
                       **kwargs) -> Any:

        _prefect_flow = state

        return _prefect_flow(*args, **kwargs)


# TODO: Refactor redundant code



# generated by datamodel-codegen:
#   filename:  ontology_annotation_schema.json
#   timestamp: 2024-02-07T08:09:26+00:00

from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, Extra, Field

from omnipy.data.model import Model

from . import comment_schema


class FieldType(Enum):
    OntologyAnnotation = 'OntologyAnnotation'


class IsaOntologyReferenceSchema(BaseModel):
    class Config:
        extra = Extra.forbid
        use_enum_values = True

    field_id: Optional[str] = Field(None, alias='@id')
    field_context: Optional[str] = Field(None, alias='@context')
    field_type: Optional[FieldType] = Field(None, alias='@type')
    annotationValue: Optional[Union[str, float]] = None
    termSource: Optional[str] = Field(
        None,
        description='The abbreviated ontology name. It should correspond to one of the sources as '
        'specified in the ontologySourceReference section of the Investigation.',
    )
    termAccession: Optional[str] = None
    comments: Optional[List[comment_schema.IsaCommentModel]] = None


class IsaOntologyReferenceModel(Model[IsaOntologyReferenceSchema]):
    ...



import os
from types import ModuleType

from pydantic.utils import lenient_isinstance, lenient_issubclass

import omnipy

# TODO: Finish implementation of dynamic __all__ generation. Possibly useful together with Poe the
#       Poet (https://poethepoet.natn.io/poetry_plugin.html) for generating a fixed __all__ list as
#       part of the build process. The goal for this functionality is to follow DRY principles and
#       make sure new datasets, models, and jobs are exported. It is important to also allow export
#       under development, either through running `poetry build` or through a
#       `if not typing.TYPE_CHECKING` block or similar solution to allow dynamic __all__ generation
#       for development (with omnipy imported through `pip install -e {PATH_TO_OMNIPY}`). The exact
#       implementation is not yet decided, but the goal is to make the code the single source of
#       truth for the exported elements of omnipy.

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
__all__: list[str] = []

_all_element_names: set[str] = set()

_exclude_modules: set[str] = {
    '_dynamic_all',
    'modules.frozen',  # Recursive frozen models crashes mypy v1.10 + waiting for pydantic support
    'modules.fairtracks',
    'util.tabulate',
}
_exclude_attrs: set[str] = {
    'JobMixin',
    'JobTemplateMixin',
    'ListOfNestedStrModel',
    'SplitItemsToSubitemsModel',
    'JoinSubitemsToItemsModel',
    'QueryParamsJoinerModel',
    'QueryParamsSplitterModel',
}

_all_modules: dict[str, ModuleType] = {}

if not __all__:
    from .compute.flow import (DagFlow,
                               DagFlowTemplate,
                               FuncFlow,
                               FuncFlowTemplate,
                               LinearFlow,
                               LinearFlowTemplate)
    from .compute.job import JobTemplateMixin
    from .compute.task import Task, TaskTemplate
    from .data.dataset import Dataset, MultiModelDataset
    from .data.model import Model
    from .data.param import (bind_adjust_dataset_func,
                             bind_adjust_model_func,
                             params_dataclass,
                             ParamsBase)
    from .hub.runtime import runtime
    from .util.contexts import print_exception
    from .util.helpers import recursive_module_import_new

    __all__ = [
        'DagFlow',
        'DagFlowTemplate',
        'FuncFlow',
        'FuncFlowTemplate',
        'LinearFlow',
        'LinearFlowTemplate',
        'Task',
        'TaskTemplate',
        'Dataset',
        'MultiModelDataset',
        'Model',
        'bind_adjust_model_func',
        'bind_adjust_dataset_func',
        'params_dataclass',
        'ParamsBase',
        'runtime',
        'print_exception',
    ]
    _all_element_names = set(__all__)

    _omnipy_all = set(omnipy.__all__)

    recursive_module_import_new([ROOT_DIR], _all_modules, _exclude_modules)

    for module_name, module in _all_modules.items():
        for attr in dir(module):
            if not attr.startswith('_') and attr not in _all_element_names:
                val = getattr(module, attr)
                if attr in _exclude_attrs:
                    continue

                if lenient_issubclass(val, Model) \
                        or lenient_issubclass(val, Dataset) \
                        or lenient_isinstance(val, JobTemplateMixin):
                    print(f'Adding {attr}')
                    _all_element_names.add(attr)
                    globals()[attr] = val
                    __all__.append(attr)

    print(f'Missing elements in omnipy.__init__(): {_all_element_names - _omnipy_all}')
    print(f'Missing elements in hardcoded __all__ in _dynamic_all(): '
          f'{_omnipy_all - _all_element_names}')

    del JobTemplateMixin



__version__ = '0.17.2'

import os

from omnipy.compute.flow import (DagFlow,
                                 DagFlowTemplate,
                                 FuncFlow,
                                 FuncFlowTemplate,
                                 LinearFlow,
                                 LinearFlowTemplate)
from omnipy.compute.task import Task, TaskTemplate
from omnipy.data.dataset import Dataset, MultiModelDataset
from omnipy.data.model import Model
from omnipy.data.param import (bind_adjust_dataset_func,
                               bind_adjust_model_func,
                               params_dataclass,
                               ParamsBase)
from omnipy.hub.runtime import runtime
from omnipy.modules.general.models import (Chain2,
                                           Chain3,
                                           Chain4,
                                           Chain5,
                                           Chain6,
                                           NotIterableExceptStrOrBytesModel)
from omnipy.modules.general.tasks import convert_dataset, import_directory, split_dataset
from omnipy.modules.isa import (flatten_isa_json,
                                FlattenedIsaJsonDataset,
                                FlattenedIsaJsonModel,
                                IsaJsonDataset,
                                IsaJsonModel)
from omnipy.modules.isa.models import IsaInvestigationModel, IsaTopLevelModel
from omnipy.modules.isa.models.assay_schema import IsaAssayJsonModel
from omnipy.modules.isa.models.comment_schema import IsaCommentModel
from omnipy.modules.isa.models.data_schema import IsaDataModel
from omnipy.modules.isa.models.factor_schema import IsaFactorModel
from omnipy.modules.isa.models.factor_value_schema import IsaFactorValueModel
from omnipy.modules.isa.models.material_attribute_schema import IsaMaterialAttributeModel
from omnipy.modules.isa.models.material_attribute_value_schema import IsaMaterialAttributeValueModel
from omnipy.modules.isa.models.material_schema import IsaMaterialModel
from omnipy.modules.isa.models.ontology_annotation_schema import IsaOntologyReferenceModel
from omnipy.modules.isa.models.ontology_source_reference_schema import \
    IsaOntologySourceReferenceModel
from omnipy.modules.isa.models.organization_schema import IsaOrganizationModel
from omnipy.modules.isa.models.person_schema import IsaPersonModel
from omnipy.modules.isa.models.process_parameter_value_schema import IsaProcessParameterValueModel
from omnipy.modules.isa.models.process_schema import IsaProcessOrProtocolApplicationModel
from omnipy.modules.isa.models.protocol_parameter_schema import IsaProtocolParameterModel
from omnipy.modules.isa.models.protocol_schema import IsaProtocolModel
from omnipy.modules.isa.models.publication_schema import IsaPublicationModel
from omnipy.modules.isa.models.sample_schema import IsaSampleModel
from omnipy.modules.isa.models.source_schema import IsaSourceModel
from omnipy.modules.isa.models.study_group import IsaStudyGroupModel
from omnipy.modules.isa.models.study_schema import IsaStudyModel
from omnipy.modules.json.datasets import (JsonDataset,
                                          JsonDictDataset,
                                          JsonDictOfDictsDataset,
                                          JsonDictOfDictsOfScalarsDataset,
                                          JsonDictOfListsDataset,
                                          JsonDictOfListsOfDictsDataset,
                                          JsonDictOfListsOfScalarsDataset,
                                          JsonDictOfNestedListsDataset,
                                          JsonDictOfScalarsDataset,
                                          JsonListDataset,
                                          JsonListOfDictsDataset,
                                          JsonListOfDictsOfScalarsDataset,
                                          JsonListOfListsDataset,
                                          JsonListOfListsOfScalarsDataset,
                                          JsonListOfNestedDictsDataset,
                                          JsonListOfScalarsDataset,
                                          JsonNestedDictsDataset,
                                          JsonNestedListsDataset,
                                          JsonOnlyDictsDataset,
                                          JsonOnlyListsDataset,
                                          JsonScalarDataset)
from omnipy.modules.json.flows import (flatten_nested_json,
                                       transpose_dict_of_dicts_2_list_of_dicts,
                                       transpose_dicts_of_lists_of_dicts_2_lists_of_dicts)
from omnipy.modules.json.models import (JsonCustomDictModel,
                                        JsonCustomListModel,
                                        JsonDictModel,
                                        JsonDictOfDictsModel,
                                        JsonDictOfDictsOfScalarsModel,
                                        JsonDictOfListsModel,
                                        JsonDictOfListsOfDictsModel,
                                        JsonDictOfListsOfScalarsModel,
                                        JsonDictOfNestedListsModel,
                                        JsonDictOfScalarsModel,
                                        JsonListModel,
                                        JsonListOfDictsModel,
                                        JsonListOfDictsOfScalarsModel,
                                        JsonListOfListsModel,
                                        JsonListOfListsOfScalarsModel,
                                        JsonListOfNestedDictsModel,
                                        JsonListOfScalarsModel,
                                        JsonModel,
                                        JsonNestedDictsModel,
                                        JsonNestedListsModel,
                                        JsonOnlyDictsModel,
                                        JsonOnlyListsModel,
                                        JsonScalarModel)
from omnipy.modules.json.tasks import convert_dataset_string_to_json, transpose_dicts_2_lists
from omnipy.modules.pandas.models import (ListOfPandasDatasetsWithSameNumberOfFiles,
                                          PandasDataset,
                                          PandasModel)
from omnipy.modules.pandas.tasks import (cartesian_product,
                                         concat_dataframes_across_datasets,
                                         convert_dataset_csv_to_pandas,
                                         convert_dataset_list_of_dicts_to_pandas,
                                         convert_dataset_pandas_to_csv,
                                         extract_columns_as_files,
                                         join_tables)
from omnipy.modules.raw.datasets import (BytesDataset,
                                         JoinColumnsToLinesDataset,
                                         JoinItemsDataset,
                                         JoinLinesDataset,
                                         SplitLinesToColumnsDataset,
                                         SplitToItemsDataset,
                                         SplitToLinesDataset,
                                         StrDataset)
from omnipy.modules.raw.models import (BytesModel,
                                       JoinColumnsToLinesModel,
                                       JoinItemsModel,
                                       JoinLinesModel,
                                       MatchItemsModel,
                                       NestedJoinItemsModel,
                                       NestedSplitToItemsModel,
                                       SplitLinesToColumnsModel,
                                       SplitToItemsModel,
                                       SplitToLinesModel,
                                       StrModel)
from omnipy.modules.raw.tasks import (concat_all,
                                      decode_bytes,
                                      modify_all_lines,
                                      modify_datafile_contents,
                                      modify_each_line,
                                      union_all)
from omnipy.modules.remote.datasets import HttpUrlDataset
from omnipy.modules.remote.models import HttpUrlModel, QueryParamsModel, UrlPathModel
from omnipy.modules.remote.tasks import (async_load_urls_into_new_dataset,
                                         get_bytes_from_api_endpoint,
                                         get_json_from_api_endpoint,
                                         get_str_from_api_endpoint,
                                         load_urls_into_new_dataset)
from omnipy.modules.tables.datasets import (TableDictOfDictsOfJsonScalarsDataset,
                                            TableDictOfListsOfJsonScalarsDataset,
                                            TableListOfDictsOfJsonScalarsDataset,
                                            TableListOfListsOfJsonScalarsDataset,
                                            TableOfPydanticRecordsDataset,
                                            TableWithColNamesDataset)
from omnipy.modules.tables.models import (PydanticRecordModel,
                                          TableDictOfDictsOfJsonScalarsModel,
                                          TableDictOfListsOfJsonScalarsModel,
                                          TableListOfDictsOfJsonScalarsModel,
                                          TableListOfListsOfJsonScalarsModel,
                                          TableOfPydanticRecordsModel,
                                          TableWithColNamesModel)
from omnipy.modules.tables.tasks import (create_row_index_from_column,
                                         remove_columns,
                                         rename_col_names,
                                         transpose_columns_with_data_files)
from omnipy.util.contexts import print_exception

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

__all__ = [
    'runtime',
    'DagFlow',
    'DagFlowTemplate',
    'FuncFlow',
    'FuncFlowTemplate',
    'LinearFlow',
    'LinearFlowTemplate',
    'Task',
    'TaskTemplate',
    'Dataset',
    'MultiModelDataset',
    'Model',
    'FlattenedIsaJsonDataset',
    'FlattenedIsaJsonModel',
    'IsaJsonModel',
    'IsaJsonDataset',
    'IsaInvestigationModel',
    'IsaTopLevelModel',
    'IsaAssayJsonModel',
    'IsaCommentModel',
    'IsaDataModel',
    'IsaFactorModel',
    'IsaFactorValueModel',
    'IsaMaterialAttributeModel',
    'IsaMaterialAttributeValueModel',
    'IsaMaterialModel',
    'IsaOntologyReferenceModel',
    'IsaOntologySourceReferenceModel',
    'IsaOrganizationModel',
    'IsaPersonModel',
    'IsaProcessParameterValueModel',
    'IsaProcessOrProtocolApplicationModel',
    'IsaProtocolParameterModel',
    'IsaProtocolModel',
    'IsaPublicationModel',
    'IsaSampleModel',
    'IsaSourceModel',
    'IsaStudyGroupModel',
    'IsaStudyModel',
    'JsonCustomDictModel',
    'JsonCustomListModel',
    'JsonDataset',
    'JsonDictDataset',
    'JsonDictOfDictsDataset',
    'JsonDictOfDictsOfScalarsDataset',
    'JsonDictOfListsDataset',
    'JsonDictOfListsOfDictsDataset',
    'JsonDictOfListsOfScalarsDataset',
    'JsonDictOfNestedListsDataset',
    'JsonDictOfScalarsDataset',
    'JsonListDataset',
    'JsonListOfDictsDataset',
    'JsonListOfDictsOfScalarsDataset',
    'JsonListOfListsDataset',
    'JsonListOfListsOfScalarsDataset',
    'JsonListOfNestedDictsDataset',
    'JsonListOfScalarsDataset',
    'JsonNestedDictsDataset',
    'JsonNestedListsDataset',
    'JsonOnlyDictsDataset',
    'JsonOnlyListsDataset',
    'JsonScalarDataset',
    'JsonDictModel',
    'JsonDictOfDictsModel',
    'JsonDictOfDictsOfScalarsModel',
    'JsonDictOfListsModel',
    'JsonDictOfListsOfDictsModel',
    'JsonDictOfListsOfScalarsModel',
    'JsonDictOfNestedListsModel',
    'JsonDictOfScalarsModel',
    'JsonListModel',
    'JsonListOfDictsModel',
    'JsonListOfDictsOfScalarsModel',
    'JsonListOfListsModel',
    'JsonListOfListsOfScalarsModel',
    'JsonListOfNestedDictsModel',
    'JsonListOfScalarsModel',
    'JsonModel',
    'JsonNestedDictsModel',
    'JsonNestedListsModel',
    'JsonOnlyDictsModel',
    'JsonOnlyListsModel',
    'JsonScalarModel',
    'ListOfPandasDatasetsWithSameNumberOfFiles',
    'PandasModel',
    'PandasDataset',
    'BytesDataset',
    'SplitToLinesDataset',
    'JoinLinesDataset',
    'SplitToItemsDataset',
    'JoinItemsDataset',
    'SplitLinesToColumnsDataset',
    'JoinColumnsToLinesDataset',
    'StrDataset',
    'BytesModel',
    'SplitToLinesModel',
    'JoinLinesModel',
    'SplitToItemsModel',
    'JoinItemsModel',
    'SplitLinesToColumnsModel',
    'JoinColumnsToLinesModel',
    'NestedJoinItemsModel',
    'NestedSplitToItemsModel',
    'MatchItemsModel',
    'StrModel',
    'TableDictOfDictsOfJsonScalarsDataset',
    'TableDictOfListsOfJsonScalarsDataset',
    'TableListOfListsOfJsonScalarsDataset',
    'TableListOfDictsOfJsonScalarsDataset',
    'TableOfPydanticRecordsDataset',
    'TableWithColNamesDataset',
    'PydanticRecordModel',
    'TableOfPydanticRecordsModel',
    'TableWithColNamesModel',
    'TableDictOfDictsOfJsonScalarsModel',
    'TableDictOfListsOfJsonScalarsModel',
    'TableListOfDictsOfJsonScalarsModel',
    'TableListOfListsOfJsonScalarsModel',
    'bind_adjust_model_func',
    'bind_adjust_dataset_func',
    'params_dataclass',
    'ParamsBase',
    'NotIterableExceptStrOrBytesModel',
    'Chain2',
    'Chain3',
    'Chain4',
    'Chain5',
    'Chain6',
    'import_directory',
    'split_dataset',
    'convert_dataset',
    'flatten_isa_json',
    'flatten_nested_json',
    'convert_dataset_string_to_json',
    'transpose_dicts_2_lists',
    'transpose_dict_of_dicts_2_list_of_dicts',
    'transpose_dicts_of_lists_of_dicts_2_lists_of_dicts',
    'concat_dataframes_across_datasets',
    'convert_dataset_csv_to_pandas',
    'convert_dataset_pandas_to_csv',
    'convert_dataset_list_of_dicts_to_pandas',
    'extract_columns_as_files',
    'join_tables',
    'cartesian_product',
    'decode_bytes',
    'modify_all_lines',
    'modify_datafile_contents',
    'modify_each_line',
    'concat_all',
    'union_all',
    'QueryParamsModel',
    'HttpUrlModel',
    'UrlPathModel',
    'HttpUrlDataset',
    'get_bytes_from_api_endpoint',
    'get_json_from_api_endpoint',
    'get_str_from_api_endpoint',
    'async_load_urls_into_new_dataset',
    'load_urls_into_new_dataset',
    'create_row_index_from_column',
    'remove_columns',
    'rename_col_names',
    'transpose_columns_with_data_files',
    'print_exception',
]



from typing import Any, Callable, IO, Iterator, Protocol, Type, TypeVar

from pydantic.fields import Undefined, UndefinedType

from omnipy.api.protocols.private.log import CanLog

_ModelT = TypeVar('_ModelT')


class IsDataset(Protocol[_ModelT]):
    """
    Dict-based container of data files that follow a specific Model
    """
    def __init__(
        self,
        value: dict[str, object] | Iterator[tuple[str, object]] | UndefinedType = Undefined,
        *,
        data: dict[str, object] | UndefinedType = Undefined,
        **input_data: object,
    ) -> None:
        ...

    @classmethod
    def get_model_class(cls) -> Type[_ModelT]:
        """
        Returns the concrete Model class used for all data files in the dataset, e.g.:
        `Model[list[int]]`
        :return: The concrete Model class used for all data files in the dataset
        """
        ...

    def to_data(self) -> dict[str, Any]:
        ...

    def from_data(self,
                  data: dict[str, Any] | Iterator[tuple[str, Any]],
                  update: bool = True) -> None:
        ...

    def to_json(self, pretty=True) -> dict[str, str]:
        ...

    def from_json(self,
                  data: dict[str, str] | Iterator[tuple[str, str]],
                  update: bool = True) -> None:
        ...

    @classmethod
    def to_json_schema(cls, pretty=True) -> str | dict[str, str]:
        ...

    def as_multi_model_dataset(self) -> 'IsMultiModelDataset[_ModelT]':
        ...


class IsMultiModelDataset(Protocol[_ModelT]):
    """
        Variant of Dataset that allows custom models to be set on individual data files
    """
    def set_model(self, data_file: str, model: _ModelT) -> None:
        ...

    def get_model(self, data_file: str) -> _ModelT:
        ...


class IsSerializer(Protocol):
    """"""
    @classmethod
    def is_dataset_directly_supported(cls, dataset: IsDataset) -> bool:
        pass

    @classmethod
    def get_dataset_cls_for_new(cls) -> Type[IsDataset]:
        pass

    @classmethod
    def get_output_file_suffix(cls) -> str:
        pass

    @classmethod
    def serialize(cls, dataset: IsDataset) -> bytes | memoryview:
        pass

    @classmethod
    def deserialize(cls, serialized: bytes, any_file_suffix=False) -> IsDataset:
        pass


class IsTarFileSerializer(IsSerializer):
    @classmethod
    def create_tarfile_from_dataset(cls,
                                    dataset: IsDataset,
                                    data_encode_func: Callable[[Any], bytes | memoryview]):
        """"""

    @classmethod
    def create_dataset_from_tarfile(cls,
                                    dataset: IsDataset,
                                    tarfile_bytes: bytes,
                                    data_decode_func: Callable[[IO[bytes]], Any],
                                    dictify_object_func: Callable[[str, Any], dict | str],
                                    import_method='from_data',
                                    any_file_suffix: bool = False):
        ...


class IsSerializerRegistry(Protocol):
    """"""
    def __init__(self) -> None:
        ...

    def register(self, serializer_cls: Type[IsSerializer]) -> None:
        ...

    @property
    def serializers(self) -> tuple[Type[IsSerializer], ...]:
        ...

    @property
    def tar_file_serializers(self) -> tuple[Type[IsTarFileSerializer], ...]:
        ...

    def auto_detect(self, dataset: IsDataset):
        ...

    def auto_detect_tar_file_serializer(self, dataset: IsDataset):
        ...

    @classmethod
    def _autodetect_serializer(cls, dataset, serializers):
        ...

    def detect_tar_file_serializers_from_dataset_cls(self, dataset: IsDataset):
        ...

    def detect_tar_file_serializers_from_file_suffix(self, file_suffix: str):
        ...

    def load_from_tar_file_path_based_on_file_suffix(self,
                                                     log_obj: CanLog,
                                                     tar_file_path: str,
                                                     to_dataset: IsDataset):
        ...

    def load_from_tar_file_path_based_on_dataset_cls(self,
                                                     log_obj: CanLog,
                                                     tar_file_path: str,
                                                     to_dataset: IsDataset):
        ...



from copy import copy, deepcopy
from typing import cast

from omnipy.compute.task import TaskTemplate
from omnipy.compute.typing import mypy_fix_task_template

from ..json.datasets import JsonListOfDictsDataset
from ..json.typedefs import JsonScalar
from .datasets import TableWithColNamesDataset
from .models import (TableDictOfDictsOfJsonScalarsModel,
                     TableListOfDictsOfJsonScalarsModel,
                     TableWithColNamesModel)


@mypy_fix_task_template
@TaskTemplate()
def remove_columns(json_dataset: JsonListOfDictsDataset,
                   column_keys_for_data_items: dict[str, list[str]]) -> JsonListOfDictsDataset:
    # TODO: implement general solution to make sure that one does not modify input data by
    #       automatically copying or otherwise. Perhaps setting immutable/frozen option in pydantic,
    #       if available?
    #
    dataset_copy = deepcopy(json_dataset)
    for data_item_key, column_keys in column_keys_for_data_items.items():
        for record in dataset_copy[data_item_key]:
            for column in column_keys:
                if column in record:
                    del record[column]
    return JsonListOfDictsDataset(json_dataset.to_data())


@mypy_fix_task_template
@TaskTemplate(iterate_over_data_files=True, output_dataset_cls=TableWithColNamesDataset)
def rename_col_names(data_file: TableWithColNamesModel,
                     prev2new_keymap: dict[str, str]) -> TableWithColNamesModel:
    return TableWithColNamesModel([{
        prev2new_keymap[key] if key in prev2new_keymap else key: val for key, val in row.items()
    } for row in data_file])


@mypy_fix_task_template
@TaskTemplate()
def transpose_columns_with_data_files(dataset: TableWithColNamesDataset,
                                      exclude_cols: tuple[str]) -> None:
    output_dataset = JsonListOfDictsDataset()

    max_len = max(len(data_file) for data_file in dataset.values())

    # TODO: Make Dataset behave like a defaultDict, possibly also with auto-expanding lists?
    for column_name in dataset.col_names:
        if column_name not in exclude_cols:
            output_dataset[column_name] = [{}] * max_len

    for data_file_name, data_file in dataset.items():
        for row_i, el in enumerate(data_file):
            for key, val in el.items():
                if key in exclude_cols:
                    for data_file in output_dataset.values():
                        data_file[row_i][key] = val
                else:
                    output_dataset[key][row_i][data_file_name] = val

    return output_dataset


# @mypy_fix_task_template
# @TaskTemplate()
# def convert_to_1nf(input_dataset: Dataset[TableOfStringsAndLists]) -> Dataset[TableOfStrings]:
#     out_dataset = Dataset[TableOfStrings]()
#     for table_name, table in input_dataset.items():
#         # for record in table:
#         #     ...
#         out_dataset[table_name] = table
#     return out_dataset

# TODO: Somehow fix so that we do not need to call Model.contents within a task


@TaskTemplate(iterate_over_data_files=True)
def create_row_index_from_column(list_of_dicts: TableListOfDictsOfJsonScalarsModel,
                                 column_key: str) -> TableDictOfDictsOfJsonScalarsModel:
    output_dict = {}
    input_table = cast(list[dict[str, JsonScalar]], list_of_dicts.to_data())
    for item in input_table:
        item_copy = copy(item)
        new_key = item[column_key]
        del item_copy[column_key]
        output_dict[new_key] = item_copy
    return TableDictOfDictsOfJsonScalarsModel(output_dict)



from typing import cast

from omnipy.api.protocols.private.compute.job import IsJobBase
from omnipy.compute.mixins.name import NameJobBaseMixin


class ResultKeyFuncJobBaseMixin:
    def __init__(self, *, result_key: str | None = None):
        self_as_name_job_base_mixin = cast(NameJobBaseMixin, self)

        self._result_key = result_key

        if self.result_key is not None:
            self_as_name_job_base_mixin._check_not_empty_string('result_key', self.result_key)

    @property
    def result_key(self) -> str | None:
        return self._result_key

    def _call_job(self, *args: object, **kwargs: object) -> object:
        super_as_job_base = cast(IsJobBase, super())
        result = super_as_job_base._call_job(*args, **kwargs)

        if self._result_key:
            return {self._result_key: result}
        else:
            return result



import inspect
import os
from pathlib import Path
import sys


def set_prefect_config_path():
    prefect_module_dir = os.path.dirname(inspect.getabsfile(inspect.currentframe()))

    os.environ['PREFECT_LOGGING_SETTINGS_PATH'] = \
        f"{Path(prefect_module_dir).joinpath('settings', 'logging.yml')}"


def use_local_api_for_tests():
    if 'pytest' in sys.modules:
        os.environ['PREFECT_API_KEY'] = ''
        os.environ['PREFECT_API_URL'] = ''


set_prefect_config_path()
use_local_api_for_tests()

from prefect import Flow as PrefectFlow  # noqa
from prefect import flow as prefect_flow  # noqa
from prefect import State  # noqa
from prefect import Task as PrefectTask  # noqa
from prefect import task as prefect_task  # noqa
from prefect.tasks import task_input_hash  # noqa
from prefect.utilities.names import generate_slug  # noqa



from omnipy.api.protocols.public.data import IsSerializerRegistry


def register_serializers(registry: IsSerializerRegistry):
    from .json.serializers import JsonDatasetToTarFileSerializer
    from .pandas.serializers import PandasDatasetToTarFileSerializer
    from .raw.serializers import (RawBytesDatasetToTarFileSerializer,
                                  RawStrDatasetToTarFileSerializer)

    registry.register(RawStrDatasetToTarFileSerializer)
    registry.register(RawBytesDatasetToTarFileSerializer)
    registry.register(JsonDatasetToTarFileSerializer)
    registry.register(PandasDatasetToTarFileSerializer)


def get_serializer_registry():
    from omnipy.data.serializer import SerializerRegistry
    from omnipy.hub.runtime import runtime

    serializer_registry = SerializerRegistry() if runtime is None else \
        runtime.objects.serializers
    if len(serializer_registry.serializers) == 0:
        register_serializers(serializer_registry)
    return serializer_registry


# TODO: Add module with helper classes/functions/takss to make it simpler to contact REST apis
#       Augmentation service should have some attempts at this.



from typing import Any, cast, get_args

from typing_extensions import TypeVar

from omnipy.api.exceptions import FailedDataError, PendingDataError
from omnipy.api.protocols.public.data import IsDataset
from omnipy.api.typedefs import TypeForm
from omnipy.data.helpers import FailedData, HasData, is_model_subclass, PendingData
from omnipy.util.decorators import call_super_if_available
from omnipy.util.helpers import is_union

# ModelT = TypeVar('ModelT', bound=IsModel)
ModelT = TypeVar('ModelT')
T = TypeVar('T')


class TaskDatasetMixin:
    @call_super_if_available(call_super_before_method=True)
    @classmethod
    def _prepare_params(cls, params: TypeForm) -> TypeForm:
        cleaned_params = cls._clean_model(params)
        if is_model_subclass(cleaned_params):
            return cleaned_params | PendingData | FailedData
        else:
            return cleaned_params

    @call_super_if_available(call_super_before_method=True)
    @classmethod
    def _clean_model(cls, model: TypeForm) -> TypeForm:
        args = get_args(model)
        if is_union(model) and len(args) == 3 and args[1:] == (PendingData, FailedData):
            return args[0]
        else:
            return model

    @property
    def available_data(self) -> IsDataset[type[ModelT]]:
        self_with_data = cast(HasData, self)
        copy = cast(HasData, self.__class__())
        copy.data = {
            key: val for key,
            val in self_with_data.data.items() if not isinstance(val, (PendingData, FailedData))
        }
        return cast(IsDataset[type[ModelT]], copy)

    @property
    def pending_data(self) -> IsDataset[type[ModelT]]:
        self_with_data = cast(HasData, self)
        copy = cast(HasData, self.__class__())
        copy.data = {
            key: val for key, val in self_with_data.data.items() if isinstance(val, PendingData)
        }
        return cast(IsDataset[type[ModelT]], copy)

    @property
    def failed_data(self) -> IsDataset[type[ModelT]]:
        self_with_data = cast(HasData, self)
        copy = cast(HasData, self.__class__())
        copy.data = {
            key: val for key, val in self_with_data.data.items() if isinstance(val, FailedData)
        }
        return cast(IsDataset[type[ModelT]], copy)

    def pending_task_details(self) -> dict[str, PendingData]:
        self_with_data = cast(HasData, self)
        return {
            key: val for key, val in self_with_data.data.items() if isinstance(val, PendingData)
        }

    def failed_task_details(self) -> dict[str, FailedData]:
        self_with_data = cast(HasData, self)
        return {key: val for key, val in self_with_data.data.items() if isinstance(val, FailedData)}

    @call_super_if_available(call_super_before_method=True)
    def _check_value(self, value: Any) -> Any:
        if isinstance(value, PendingData):
            raise PendingDataError(f'Dataset is still awaiting data from job "{value.job_name}"')
        elif isinstance(value, FailedData):
            raise FailedDataError(
                f'Job "{value.job_name}" failed to return data: {repr(value.exception)}'
            ) from value.exception

        return value



from datetime import date, datetime
import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path


class DailyRotatingFileHandler(RotatingFileHandler):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._last_log_date_file_path: str = str(
            Path(self.baseFilename).parent / '_last_log_date.txt')
        self._last_log_date: date | None = self._read_log_date()

    @staticmethod
    def _get_date_from_created(record: logging.LogRecord) -> date:
        return datetime.fromtimestamp(record.created).date()

    def _store_log_date(self, log_date: date) -> None:
        self._last_log_date = log_date
        with open(self._last_log_date_file_path, 'w') as cur_date_file:
            cur_date_file.write(self._last_log_date.isoformat())

    def _read_log_date(self) -> date | None:
        try:
            with open(self._last_log_date_file_path, 'r') as cur_date_file:
                return date.fromisoformat(cur_date_file.read())
        except FileNotFoundError:
            return None

    def shouldRollover(self, record: logging.LogRecord) -> bool:
        if super().shouldRollover(record):
            return True

        log_date = self._get_date_from_created(record)
        if log_date != self._last_log_date:
            do_rollover = self._last_log_date is not None
            self._store_log_date(log_date)
            return do_rollover
        return False



from typing import Callable, cast

from omnipy.api.protocols.private.compute.job import (IsJob,
                                                      IsJobTemplate,
                                                      IsTaskTemplateArgsJobTemplate)
from omnipy.api.protocols.public.compute import IsTaskTemplate
from omnipy.compute.func_job import FuncArgJobBase

# TODO: Update TaskTemplateArgsJobBase typing to also allow sub-flows


class TaskTemplateArgsJobBase(FuncArgJobBase):
    def __init__(self, job_func: Callable, *task_templates: IsTaskTemplate,
                 **kwargs: object) -> None:
        self._task_templates: tuple[IsTaskTemplate, ...] = task_templates

    def _get_init_args(self) -> tuple[object, ...]:
        return self._job_func, *self._task_templates

    @property
    def task_templates(self) -> tuple[IsTaskTemplate, ...]:
        return self._task_templates

    def _refine(
            self,
            *task_templates: IsTaskTemplate,
            update: bool = True,
            **kwargs: object
    ) -> IsTaskTemplateArgsJobTemplate[IsTaskTemplate, IsJobTemplate, IsJob]:

        refined_template = super()._refine(
            *task_templates,
            update=update,
            **kwargs,
        )
        return cast(IsTaskTemplateArgsJobTemplate[IsTaskTemplate, IsJobTemplate, IsJob],
                    refined_template)



from dataclasses import dataclass, field
import logging
from logging import StreamHandler
from logging.handlers import RotatingFileHandler
import os

from omnipy.api.protocols.public.config import IsRootLogConfig
from omnipy.config.root_log import RootLogConfig
from omnipy.hub.log.handlers import DailyRotatingFileHandler
from omnipy.util.helpers import get_datetime_format
from omnipy.util.publisher import RuntimeEntryPublisher


@dataclass
class RootLogConfigEntryPublisher(RootLogConfig, RuntimeEntryPublisher):
    ...


@dataclass
class RootLogObjects:
    _config: IsRootLogConfig = field(
        init=False, repr=False, default_factory=RootLogConfigEntryPublisher)

    formatter: logging.Formatter | None = None
    stdout_handler: StreamHandler | None = None
    stderr_handler: StreamHandler | None = None
    file_handler: RotatingFileHandler | None = None

    def __post_init__(self):
        self._configure_all_objects()

    def set_config(self, config: IsRootLogConfig):
        self._config = config
        self._configure_all_objects()

    def _configure_all_objects(self):
        self._remove_all_handlers_from_root_logger()

        self._configure_formatter()

        self._configure_stdout_handler()
        self._configure_stderr_handler()
        self._configure_file_handler()

        filters = self._configure_common_filters()
        self._add_all_handlers_to_root_logger(filters)

    def _configure_formatter(self):
        if self._config.log_format_str:
            datetime_fmt = get_datetime_format(self._config.locale)
            self.formatter = logging.Formatter(self._config.log_format_str, datetime_fmt, style='{')
        else:
            self.formatter = None

    def _configure_stdout_handler(self):
        if self._config.log_to_stdout:
            config = self._config

            class StdErrBasedMaxLevelFilter(logging.Filter):
                def filter(self, record):
                    return record.levelno < config.stderr_log_min_level

            self.stdout_handler = StreamHandler(self._config.stdout)
            self.stdout_handler.setLevel(self._config.stdout_log_min_level)
            if self._config.log_to_stderr:
                self.stdout_handler.addFilter(StdErrBasedMaxLevelFilter())
        else:
            self.stdout_handler = None

    def _configure_stderr_handler(self) -> None:
        if self._config.log_to_stderr:
            self.stderr_handler = StreamHandler(self._config.stderr)
            self.stderr_handler.setLevel(self._config.stderr_log_min_level)
        else:
            self.stderr_handler = None

    def _configure_file_handler(self) -> None:
        if self._config.log_to_file:
            log_file_path = self._config.file_log_path
            log_dir_path = os.path.dirname(log_file_path)
            if not os.path.exists(log_dir_path):
                os.makedirs(log_dir_path)

            self.file_handler = DailyRotatingFileHandler(log_file_path, backupCount=7)
            self.file_handler.setLevel(self._config.file_log_min_level)
        else:
            self.file_handler = None

    def _configure_common_filters(self):
        class ExtractEngineFilter(logging.Filter):
            def filter(self, record):
                record.engine = f"{record.name.split('.')[0].upper()}"
                if len(record.engine) < 7:
                    record.engine += ' '
                return True

        class SetTimestampFilter(logging.Filter):
            def filter(self, record):
                if hasattr(record, 'timestamp'):
                    record.created = record.timestamp
                return True

        return [ExtractEngineFilter(), SetTimestampFilter()]

    def _remove_all_handlers_from_root_logger(self):
        root_logger = logging.root
        handlersToRemove = [
            handler for handler in root_logger.handlers
            if isinstance(handler, StreamHandler) or isinstance(handler, DailyRotatingFileHandler)
        ]
        for handler in handlersToRemove:
            root_logger.removeHandler(handler)

    def _add_handler_to_root_logger(
        self,
        handler: logging.Handler | None,
        filters: list[logging.Filter],
    ):
        if handler:
            root_logger = logging.root
            if self.formatter:
                handler.setFormatter(self.formatter)
            for filter in filters:
                handler.addFilter(filter)
            root_logger.addHandler(handler)

    def _add_all_handlers_to_root_logger(self, filters: list[logging.Filter]):
        self._add_handler_to_root_logger(self.stdout_handler, filters)
        self._add_handler_to_root_logger(self.stderr_handler, filters)
        self._add_handler_to_root_logger(self.file_handler, filters)



from typing import Generic, TypeVar

from pydantic import BaseModel

from omnipy.data.dataset import Dataset

from .models import (TableDictOfDictsOfJsonScalarsModel,
                     TableDictOfListsOfJsonScalarsModel,
                     TableListOfDictsOfJsonScalarsModel,
                     TableListOfListsOfJsonScalarsModel,
                     TableOfPydanticRecordsModel,
                     TableWithColNamesModel)


class TableListOfListsOfJsonScalarsDataset(Dataset[TableListOfListsOfJsonScalarsModel]):
    ...


class TableListOfDictsOfJsonScalarsDataset(Dataset[TableListOfDictsOfJsonScalarsModel]):
    ...


class TableDictOfDictsOfJsonScalarsDataset(Dataset[TableDictOfDictsOfJsonScalarsModel]):
    ...


class TableDictOfListsOfJsonScalarsDataset(Dataset[TableDictOfListsOfJsonScalarsModel]):
    ...


class TableWithColNamesDataset(Dataset[TableWithColNamesModel]):
    @property
    def col_names(self) -> tuple[str]:
        col_names = {}
        for data_file in self.values():
            col_names.update(dict.fromkeys(data_file.col_names))
        return tuple(col_names.keys())


_PydanticModelT = TypeVar('_PydanticModelT', bound=BaseModel)


class TableOfPydanticRecordsDataset(Dataset[TableOfPydanticRecordsModel[_PydanticModelT]],
                                    Generic[_PydanticModelT]):
    ...



from abc import ABCMeta, abstractmethod
import functools
import inspect
import os
from textwrap import dedent
from types import ModuleType
from typing import Any, cast

from devtools import debug, PrettyFormat
import humanize
from IPython.display import display
from IPython.lib.pretty import RepresentationPrinter
import objsize
from pydantic.generics import GenericModel

from omnipy.data.data_class_creator import DataClassBase
from omnipy.data.helpers import FailedData, PendingData
from omnipy.util.helpers import get_first_item, has_items, is_non_str_byte_iterable
from omnipy.util.tabulate import tabulate


class BaseDisplayMixin(metaclass=ABCMeta):
    def _repr_pretty_(self, p: RepresentationPrinter, cycle: bool) -> None:
        if cycle:
            p.text(f'{cast(GenericModel, self).__repr_name__()}(...)')
        else:
            if len(p.stack) == 1:
                p.text(self.pretty_repr())
            else:
                p.text(self.__repr__())

    @abstractmethod
    def pretty_repr(self) -> str:
        ...

    def display(self) -> None:
        display(self)

    def __str__(self) -> str:
        return repr(self)


class ModelDisplayMixin(BaseDisplayMixin):
    def pretty_repr(self) -> str:
        from omnipy.data.dataset import Dataset, Model

        outer_type = cast(Model, self).outer_type()
        if inspect.isclass(outer_type) and issubclass(outer_type, Dataset):
            return cast(Dataset, self.contents).pretty_repr()

        # tabulate.PRESERVE_WHITESPACE = True  # Does not seem to work together with 'maxcolwidths'

        terminal_size_cols = cast(DataClassBase, self).config.terminal_size_columns
        terminal_size_lines = cast(DataClassBase, self).config.terminal_size_lines

        header_column_width = len('(bottom')
        num_columns = 2
        table_chars_width = 3 * num_columns + 1
        data_column_width = terminal_size_cols - table_chars_width - header_column_width

        data_indent = 2
        extra_space_due_to_escaped_chars = 12

        debug_module = cast(ModuleType, inspect.getmodule(debug))
        debug_module.pformat = PrettyFormat(  # type: ignore[attr-defined]
            indent_step=data_indent,
            simple_cutoff=20,
            width=data_column_width - data_indent - extra_space_due_to_escaped_chars,
            yield_from_generators=True,
        )

        structure = str(debug.format(self))
        structure_lines = structure.splitlines()
        new_structure_lines = dedent(os.linesep.join(structure_lines[1:])).splitlines()
        if new_structure_lines[0].startswith('self: '):
            new_structure_lines[0] = new_structure_lines[0][5:]
        max_section_height = (terminal_size_lines - 8) // 2
        structure_len = len(new_structure_lines)

        def _is_table():
            data = self.to_data()
            if is_non_str_byte_iterable(data) and has_items(data):
                first_level_item = get_first_item(data)
                if is_non_str_byte_iterable(first_level_item) and has_items(first_level_item):
                    second_level_item = get_first_item(first_level_item)
                    if not is_non_str_byte_iterable(second_level_item):
                        return True
            return False

        if structure_len > max_section_height * 2 + 1:
            top_structure_end = max_section_height
            bottom_structure_start = structure_len - max_section_height

            top_structure = os.linesep.join(new_structure_lines[:top_structure_end])
            bottom_structure = os.linesep.join(new_structure_lines[bottom_structure_start:])

            out = tabulate(
                (
                    ('#', self.__class__.__name__),
                    (os.linesep.join(str(i) for i in range(top_structure_end)), top_structure),
                    (os.linesep.join(str(i) for i in range(bottom_structure_start, structure_len)),
                     bottom_structure),
                ),
                maxcolwidths=[header_column_width, data_column_width],
                tablefmt='rounded_grid',
            )
        else:
            out = tabulate(
                (
                    ('#', self.__class__.__name__),
                    (os.linesep.join(str(i) for i in range(structure_len)),
                     os.linesep.join(new_structure_lines)),
                ),
                maxcolwidths=[header_column_width, data_column_width],
                tablefmt='rounded_grid',
            )

        return out


class DatasetDisplayMixin(BaseDisplayMixin):
    def pretty_repr(self) -> str:
        from omnipy.data.dataset import Dataset

        return tabulate(
            ((
                i,
                k,
                self._type_str(v),
                self._len_if_available(v),
                self._obj_size_if_available(v),
            ) for i, (k, v) in enumerate(cast(Dataset, self).data.items())),
            ('#', 'Data file name', 'Type', 'Length', 'Size (in memory)'),
            tablefmt='rounded_outline',
        )

    @classmethod
    def _type_str(cls, obj: Any) -> str:
        if isinstance(obj, PendingData):
            return f'{obj.job_name} -> Data pending...'
        elif isinstance(obj, FailedData):
            return f'{obj.job_name} -> {obj.exception.__class__.__name__}: {obj.exception}'
        else:
            return type(obj).__name__

    @classmethod
    def _len_if_available(cls, obj: Any) -> int | str:
        try:
            return len(obj)
        except TypeError:
            return 'N/A'

    @classmethod
    def _obj_size_if_available(cls, obj: Any) -> str:
        if isinstance(obj, PendingData):
            return 'N/A'
        else:
            try:
                cached_obj_size_func = functools.cache(cls._obj_size)
                return cached_obj_size_func(obj)
            except TypeError:
                return cls._obj_size(obj)

    @staticmethod
    def _obj_size(obj: Any) -> str:
        return humanize.naturalsize(objsize.get_deep_size(obj))



# generated by datamodel-codegen:
#   filename:  sample_schema.json
#   timestamp: 2024-02-07T08:09:26+00:00

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Extra, Field

from omnipy.data.model import Model

from . import comment_schema, factor_value_schema, material_attribute_value_schema, source_schema


class FieldType(Enum):
    Sample = 'Sample'


class IsaSampleSchema(BaseModel):
    class Config:
        extra = Extra.forbid
        use_enum_values = True

    field_id: Optional[str] = Field(None, alias='@id')
    field_context: Optional[str] = Field(None, alias='@context')
    field_type: Optional[FieldType] = Field(None, alias='@type')
    name: Optional[str] = None
    characteristics: Optional[List[
        material_attribute_value_schema.IsaMaterialAttributeValueModel]] = None
    factorValues: Optional[List[factor_value_schema.IsaFactorValueModel]] = None
    derivesFrom: Optional[List[source_schema.IsaSourceModel]] = None
    comments: Optional[List[comment_schema.IsaCommentModel]] = None


class IsaSampleModel(Model[IsaSampleSchema]):
    ...



from collections import UserDict
from types import MappingProxyType
from typing import Generic, Hashable, Sequence

from typing_extensions import TypeVar

from ..general.models import NotIterableExceptStrOrBytesModel

KeyT = TypeVar('KeyT', default=str | Hashable)
ValT = TypeVar(
    'ValT', bound=NotIterableExceptStrOrBytesModel, default=NotIterableExceptStrOrBytesModel)


# Unfortunately, MappingProxyType is declared as a final class, which means it cannot be subclassed.
# Inheriting from UserDict is a workaround.
class FrozenDict(UserDict[KeyT, ValT], Generic[KeyT, ValT]):
    """
    FrozenDict works exactly like a dict except that it cannot be modified after initialisation:

        >>> from omnipy import FrozenDict, print_exception
        >>> from types import MappingProxyType
        >>>
        >>> my_dict = FrozenDict({'temperature': 'chilly'})
        >>> my_dict
        mappingproxy({'temperature': 'chilly'})
        >>> my_dict['temperature']
        'chilly'
        >>> with print_exception:
        ...     my_dict['temperature'] = 'scorching'
        TypeError: 'mappingproxy' object does not support item assignment

    In contrast to types.MappingProxyType, FrozenDict can be initialized without parameters:

        >>> with print_exception:
        ...     MappingProxyType()
        TypeError: mappingproxy() missing required argument 'mapping' (pos 1)
        >>> FrozenDict()
        mappingproxy({})

    Typically, this would be a rather useless thing to do, as the resulting object would be an empty
    dictionary that cannot be changed. With Omnipy it actually makes sense, though, as the Model
    class calls the type argument without parameters to determine the default value of the model.

    Note:
        Mutable objects within a FrozenDict are still mutable:

        >>> my_dict = FrozenDict({'temp_map': {'cold': 'chilly', 'hot': 'scorching'}})
        >>> my_dict['temp_map']['hot'] = 'blazing'
        >>> my_dict
        mappingproxy({'temp_map': {'cold': 'chilly', 'hot': 'blazing'}})

        To implement a fully immutable nested data structure, one would need a mechanism that makes
        sure only immutable types are allowed within a FrozenDict (or a tuple, which is the
        immutable list counterpart). Luckily, Omnipy provides just that guarantee with
        NestedFrozenDictsModel, NestedTuplesModel and NestedFrozenCollectionsModel.
    """
    def __init__(self,
                 unfrozen_dict: dict[KeyT, ValT] | Sequence[tuple[KeyT, ValT]] | None = None,
                 /,
                 **kwargs):
        super().__init__(unfrozen_dict, **kwargs)
        self.data: MappingProxyType[KeyT, ValT] = MappingProxyType(self.data)  # type: ignore

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data if hasattr(self, 'data') else ''})"



from copy import deepcopy
from dataclasses import dataclass
import functools
from typing import Any, Callable, cast, Concatenate, ParamSpec

from pydantic import BaseModel
from pydantic.main import create_model, ModelMetaclass, validate_model
from typing_extensions import dataclass_transform, TypeVar

ModelT = TypeVar('ModelT')
DatasetT = TypeVar('DatasetT')
ParamsT = TypeVar('ParamsT')
ParamsP = ParamSpec('ParamsP')


class ParamsMeta(ModelMetaclass):
    def __init__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, object]) -> None:
        super().__init__(name, bases, namespace)

        model_cls = cast(type[BaseModel], cls)
        default_vals = {}
        for field in model_cls.__fields__.values():
            if field.default_factory is not None:
                raise ValueError('Default factory is not supported for Params classes')

            default_vals[field.name] = field.get_default()
            if default_vals[field.name] is None and not field.allow_none:
                raise ValueError(f'{model_cls.__name__}.{field.name} must have a default value')

        values, fields_set, validation_error = \
            validate_model(model_cls, input_data={k: v for k, v in default_vals.items()})

        if validation_error:
            raise validation_error

        for key, value in values.items():
            model_cls.__fields__[key].default = value

    def __getattr__(cls, attr: str) -> object:
        model_cls = cast(type[BaseModel], cls)
        if attr in model_cls.__fields__:
            return cls._get_param_value(attr, model_cls)
        raise AttributeError(f'{model_cls.__name__}.{attr} is not defined')

    @functools.cache
    def _get_param_value(cls, attr, model_cls):
        if model_cls.__fields__[attr].default_factory is not None:
            return model_cls.__fields__[attr].default_factory()
        else:
            return model_cls.__fields__[attr].default

    def __setattr__(cls, attr: str, value: object) -> None:
        model_cls = cast(type[BaseModel], cls)
        if attr in model_cls.__fields__:
            raise AttributeError(f'{model_cls.__name__}.{attr} is read-only')
        elif not attr.startswith('_'):
            raise AttributeError(f'{model_cls.__name__}.{attr} is not defined')
        super().__setattr__(attr, value)


class ParamsBase(BaseModel, metaclass=ParamsMeta):
    class Config:
        arbitrary_types_allowed = True
        smart_union = True

    def __new__(cls, *args: object, **kwargs: object) -> None:  # type: ignore[misc]
        raise RuntimeError(f'{cls.__name__} cannot be instantiated')

    @classmethod
    def copy_and_adjust(cls, model_name: str, **kwargs: object) -> type['ParamsBase']:
        all_field_infos = {
            field_name: deepcopy(field.field_info) for field_name, field in cls.__fields__.items()
        }

        for key, value in kwargs.items():
            all_field_infos[key].default = value

        return create_model(  # type: ignore[call-overload]
            model_name,
            __base__=ParamsBase,
            **{
                field_name: (cls.__fields__[field_name].outer_type_, field_info) for field_name,
                field_info in all_field_infos.items()
            })


def bind_adjust_model_func(
    clone_model_func: Callable[..., type[ModelT]],
    params_cls: Callable[ParamsP, Any],
) -> Callable[Concatenate[str, ParamsP], type[ModelT]]:
    def _func(model_name: str, *args: ParamsP.args, **kwargs: ParamsP.kwargs) -> type[ModelT]:
        if len(args) > 0:
            raise AttributeError(f'Positional arguments are not supported for '
                                 f'{params_cls.__module__}.{params_cls.__name__}')
        new_model_cls = clone_model_func(model_name)
        new_model_cls.Params = params_cls.copy_and_adjust(  # type: ignore[attr-defined]
            'Params',
            **kwargs,
        )
        return new_model_cls

    return _func


def bind_adjust_dataset_func(
    clone_dataset_func: Callable[..., type[DatasetT]],
    model_cls: type[ModelT],
    params_cls: Callable[ParamsP, Any],
) -> Callable[Concatenate[str, str, ParamsP], type[DatasetT]]:
    def _func(dataset_name: str, model_name: str, *args: ParamsP.args,
              **kwargs: ParamsP.kwargs) -> type[DatasetT]:
        if len(args) > 0:
            raise AttributeError(f'Positional arguments are not supported for '
                                 f'{params_cls.__module__}.{params_cls.__name__}')
        new_model_cls: type[ModelT] = cast(
            type[ModelT],
            model_cls.adjust(model_name, **kwargs),  # type: ignore[attr-defined]
        )

        new_dataset_cls = clone_dataset_func(dataset_name, new_model_cls)
        new_model_cls.Params = params_cls.copy_and_adjust(  # type: ignore[attr-defined]
            'Params',
            **kwargs,
        )
        return new_dataset_cls

    return _func


@dataclass_transform(kw_only_default=True)
def params_dataclass(cls: type[ParamsT]) -> type[ParamsT]:
    def wrap(cls):
        return dataclass(cls, kw_only=True)

    return wrap(cls)



from typing import Protocol, TypeAlias

from typing_extensions import TypeVar

from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

InputT = TypeVar('InputT', default=object)
ModelInputT = TypeVar('ModelInputT', default=Model)
ReturnT = TypeVar('ReturnT', default=object)
ModelReturnT = TypeVar('ModelReturnT', default=Model)
InputTypeT: TypeAlias = InputT | ModelInputT
InputDatasetT: TypeAlias = Dataset[Model[InputT]] | Dataset[ModelInputT]
ReturnDatasetT: TypeAlias = Dataset[Model[ReturnT]] | Dataset[ModelReturnT]


class IsIterateInnerCallable(Protocol):
    """"""
    def __call__(
        self,
        data_file: InputT | ModelInputT,
        *args: object,
        **kwargs: object,
    ) -> ReturnT | ModelReturnT:
        ...

    __name__: str



# generated by datamodel-codegen:
#   filename:  assay_schema.json
#   timestamp: 2024-02-07T08:09:26+00:00

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Extra, Field

from omnipy.data.model import Model

from . import (comment_schema,
               data_schema,
               material_attribute_schema,
               material_schema,
               ontology_annotation_schema,
               process_schema,
               sample_schema)


class FieldType(Enum):
    Assay = 'Assay'


class _Materials(BaseModel):
    samples: Optional[List[sample_schema.IsaSampleModel]] = None
    otherMaterials: Optional[List[material_schema.IsaMaterialModel]] = None


class _MaterialsModel(Model[_Materials]):
    ...


class IsaAssayJsonSchema(BaseModel):
    class Config:
        extra = Extra.forbid
        use_enum_values = True

    field_id: Optional[str] = Field(None, alias='@id')
    field_context: Optional[str] = Field(None, alias='@context')
    field_type: Optional[FieldType] = Field(None, alias='@type')
    filename: Optional[str] = None
    measurementType: Optional[ontology_annotation_schema.IsaOntologyReferenceModel] = (None)
    technologyType: Optional[ontology_annotation_schema.IsaOntologyReferenceModel] = (None)
    technologyPlatform: Optional[str] = None
    dataFiles: Optional[List[data_schema.IsaDataModel]] = None
    materials: Optional[_MaterialsModel] = None
    characteristicCategories: Optional[List[material_attribute_schema.IsaMaterialAttributeModel]] =\
        Field(
            None,
            description='List of all the characteristics categories (or material attributes) '
            'defined in the study, used to avoid duplication of their declaration '
            'when each material_attribute_value is created. ')
    unitCategories: Optional[List[ontology_annotation_schema.IsaOntologyReferenceModel]] = \
        Field(
            None,
            description='List of all the unitsdefined in the study, used to avoid duplication '
                        'of their declaration when each value is created. ')
    processSequence: Optional[List[process_schema.IsaProcessOrProtocolApplicationModel]] = None
    comments: Optional[List[comment_schema.IsaCommentModel]] = None


class IsaAssayJsonModel(Model[IsaAssayJsonSchema]):
    ...



from datetime import datetime
from logging import getLogger, INFO, Logger
import time


class LogMixin:
    def __init__(self) -> None:
        self._logger: Logger = getLogger(f'{self.__class__.__module__}.{self.__class__.__name__}')
        self._logger.setLevel(INFO)

    @property
    def logger(self) -> Logger:
        return self._logger

    def log(self, log_msg: str, level: int = INFO, datetime_obj: datetime | None = None):
        if self._logger is not None:
            create_time = datetime_obj.timestamp() if datetime_obj else time.time()
            self._logger.log(level, log_msg, extra=dict(timestamp=create_time))



from types import MappingProxyType
from typing import cast, Mapping

from omnipy.api.protocols.private.compute.job import IsJobBase
from omnipy.compute.mixins.func_signature import SignatureFuncJobBaseMixin
from omnipy.compute.mixins.name import NameJobBaseMixin
from omnipy.util.helpers import repr_max_len
from omnipy.util.param_key_mapper import ParamKeyMapper


class ParamsFuncJobBaseMixin:
    def __init__(
        self,
        *,
        fixed_params: Mapping[str, object] | None = None,
        param_key_map: Mapping[str, str] | None = None,
    ):
        self_as_signature_func_job_base_mixin = cast(SignatureFuncJobBaseMixin, self)

        self._fixed_params = dict(fixed_params) if fixed_params is not None else {}
        self._param_key_mapper = ParamKeyMapper(param_key_map if param_key_map is not None else {})

        self_as_signature_func_job_base_mixin._check_param_keys_in_func_signature(
            self.fixed_params.keys(), 'fixed_params')
        self_as_signature_func_job_base_mixin._check_param_keys_in_func_signature(
            self.param_key_map.keys(), 'param_key_map')

    @property
    def fixed_params(self) -> MappingProxyType[str, object]:
        return MappingProxyType(self._fixed_params)

    @property
    def param_key_map(self) -> MappingProxyType[str, str]:
        return MappingProxyType(self._param_key_mapper.key_map)

    def _call_job(self, *args: object, **kwargs: object) -> object:
        self_as_name_job_base_mixin = cast(NameJobBaseMixin, self)
        self_as_signature_func_job_base_mixin = cast(SignatureFuncJobBaseMixin, self)

        mapped_fixed_params = self._param_key_mapper.delete_matching_keys(
            self._fixed_params, inverse=True)
        mapped_kwargs = self._param_key_mapper.map_matching_keys_delete_inverse_matches_keep_rest(
            kwargs, inverse=True)
        try:
            if len(mapped_kwargs) < len(kwargs):
                raise TypeError('Keyword arguments {} matches parameter key map inversely.'.format(
                    tuple(set(kwargs.keys()) - set(mapped_kwargs.keys()))))

            super_as_job_base = cast(IsJobBase, super())
            result = super_as_job_base._call_job(*args, **mapped_fixed_params, **mapped_kwargs)

        except TypeError as e:
            if str(e).startswith('Incorrect job function arguments'):
                signature_values = self_as_signature_func_job_base_mixin.param_signatures.values()
                raise TypeError(f'Incorrect job function arguments for job '
                                f'"{self_as_name_job_base_mixin.name}"!\n'
                                f'Job class name: {self.__class__.__name__}\n'
                                f'Current parameter key map contents: {self.param_key_map}\n'
                                f'Positional arguments: {repr_max_len(args)}\n'
                                f'Keyword arguments: {repr_max_len(kwargs)}\n'
                                f'Mapped fixed parameters: {repr_max_len(mapped_fixed_params)}\n'
                                f'Mapped keyword arguments: {repr_max_len(mapped_kwargs)}\n'
                                f'Call function signature parameters: '
                                f'{[(str(p), p.kind) for p in signature_values]}') from e
            else:
                raise

        return result



import os
from typing import Type, TypeVar

from omnipy.compute.task import TaskTemplate
from omnipy.compute.typing import mypy_fix_task_template
from omnipy.data.dataset import Dataset
from omnipy.data.model import Model

#
# @mypy_fix_task_template
# @TaskTemplate()
# def cast_dataset(dataset: Dataset, cast_model: Callable[[], ModelT]) -> ModelT:
#     out_dataset: Dataset[ModelT] = Dataset[cast_model]()
#     for name, item in dataset.items():
#         out_dataset[name] = cast(cast_model, item.to_data())
#     return out_dataset


@mypy_fix_task_template
@TaskTemplate()
def split_dataset(
        dataset: Dataset[Model[object]],
        datafile_names_for_b: list[str]) -> tuple[Dataset[Model[object]], Dataset[Model[object]]]:
    model_cls = dataset.get_model_class()
    datafile_names_for_a = set(dataset.keys()) - set(datafile_names_for_b)
    dataset_a = Dataset[model_cls](  # type: ignore
        {
            name: dataset[name] for name in dataset.keys() if name in datafile_names_for_a
        })
    dataset_b = Dataset[model_cls](  # type: ignore
        {
            name: dataset[name] for name in dataset.keys() if name in datafile_names_for_b
        })
    return dataset_a, dataset_b


@mypy_fix_task_template
@TaskTemplate()
def import_directory(directory: str,
                     exclude_prefixes: tuple[str, ...] = ('.', '_'),
                     include_suffixes: tuple[str, ...] = (),
                     model: Type[Model] = Model[str]) -> Dataset[Model]:
    dataset = Dataset[model]()
    for import_filename in os.listdir(directory):
        if not exclude_prefixes or \
                not any(import_filename.startswith(prefix) for prefix in exclude_prefixes):
            if not include_suffixes or \
                    any(import_filename.endswith(suffix) for suffix in include_suffixes):
                with open(os.path.join(directory, import_filename)) as open_file:
                    dataset_name = '_'.join(import_filename.split('.')[:-1])
                    print(f"{import_filename} -> Dataset['{dataset_name}']")
                    dataset[dataset_name] = open_file.read()
    return dataset


_DatasetT = TypeVar('_DatasetT', bound=Dataset)


@mypy_fix_task_template
@TaskTemplate
def convert_dataset(dataset: Dataset, dataset_cls: type[_DatasetT], **kwargs: object) -> _DatasetT:
    return dataset_cls(dataset, **kwargs)



from pathlib import PurePosixPath
from typing import cast, Iterable
from urllib.parse import unquote

from pydantic import BaseModel, Field
from pydantic_core import Url

from omnipy.data.model import Model
from omnipy.util.contexts import hold_and_reset_prev_attrib_value

from ..raw.models import NestedJoinItemsModel, NestedSplitToItemsModel

QueryParamsSplitterModel = NestedSplitToItemsModel.adjust(
    'QueryParamsSplitterModel', delimiters=('&', '='))

QueryParamsJoinerModel = NestedJoinItemsModel.adjust(
    'QueryParamsJoinerModel', delimiters=('&', '='))


class QueryParamsModel(Model[dict[str, str] | tuple[tuple[str, str], ...] | tuple[str, ...] | str]):
    @classmethod
    def _parse_data(
        cls, data: dict[str, str] | tuple[tuple[str, str], ...] | tuple[str, ...] | str
    ) -> dict[str, str]:
        if isinstance(data, dict):
            return data

        params_list = QueryParamsSplitterModel(data).contents
        assert all(len(param) == 2 for param in params_list), \
            (f'Each parameter must have 2 elements only: [key, value]. '
             f'Incorrect parameter list: {params_list}')
        return dict(cast(Iterable[tuple[str, str]], params_list))

    def to_data(self) -> str:
        with hold_and_reset_prev_attrib_value(self.config,
                                              'dynamically_convert_elements_to_models'):
            self.config.dynamically_convert_elements_to_models = False
            contents_as_tuple = tuple(cast(dict[str, str], self.contents).items())
            return QueryParamsJoinerModel(contents_as_tuple).to_data()  # type: ignore[return-value]

    def __str__(self) -> str:
        return self.to_data()


class UrlPathModel(Model[PurePosixPath | str]):
    @classmethod
    def _parse_data(cls, data: PurePosixPath | str) -> PurePosixPath:
        return PurePosixPath(data) if isinstance(data, str) else data

    def to_data(self) -> str:
        return str(self.contents)

    def __str__(self) -> str:
        return str(self.contents)

    def __floordiv__(self, other):
        self.contents /= other
        return self

    def __add__(self, other):
        return UrlPathModel(str(self) + other)


DEFAULT_PORTS = {80, 443}


class UrlDataclassModel(BaseModel):
    # Mutable fields
    scheme: str
    username: str | None = None
    password: str | None = None
    host: str = 'localhost'
    port: int | None = None
    path: UrlPathModel | None = None
    query: QueryParamsModel = Field(default_factory=QueryParamsModel)
    fragment: str | None = None

    def __str__(self) -> str:
        kwargs: dict[str, str | int] = {}
        for key, val in self.dict().items():
            if val is None:
                continue
            match key:
                case 'port':
                    if val in DEFAULT_PORTS:
                        continue
                    kwargs[key] = val
                case 'path':
                    # Fix for second problem described in
                    # https://github.com/pydantic/pydantic/issues/7186#issuecomment-1912791497
                    kwargs[key] = val.lstrip('/')
                case 'query':
                    if val:
                        kwargs[key] = str(val)
                case _:
                    kwargs[key] = val

        return str(Url.build(**kwargs))  # type: ignore[arg-type]


class HttpUrlModel(Model[UrlDataclassModel | str]):
    @classmethod
    def _parse_data(cls, data: UrlDataclassModel | str) -> UrlDataclassModel:
        if data == '':
            data = 'http://localhost/'
        if data == 'https://':
            data = 'https://localhost/'
        # For validation only
        url_obj = Url(str(data) if isinstance(data, UrlDataclassModel) else data)

        parts: dict[str, str | int | None] = {}
        for key in UrlDataclassModel.__fields__.keys():
            match key:
                case 'scheme':
                    val = url_obj.scheme
                    assert val in {'http', 'https'}, f'Unsupported scheme: {val}'
                    parts[key] = val
                case 'host':
                    parts[key] = url_obj.unicode_host()
                case 'username' | 'password' | 'path' | 'query' | 'fragment':
                    val = getattr(url_obj, key)
                    parts[key] = unquote(val) if val is not None else None
                case _:
                    parts[key] = getattr(url_obj, key)

        return UrlDataclassModel(**{
            key: val  # type: ignore[arg-type]
            for key, val in parts.items()
            if val is not None
        })

    def __add__(self, other: str) -> 'HttpUrlModel':
        return HttpUrlModel(str(self) + other)

    def to_data(self) -> str:
        return str(self.contents)

    def __str__(self) -> str:
        return str(self.contents)



from dataclasses import dataclass, field
import locale as pkg_locale
import logging
from pathlib import Path
import sys
from typing import TextIO

from omnipy.api.typedefs import LocaleType


def _get_log_path() -> str:
    return str(Path.cwd() / 'logs' / 'omnipy.log')


@dataclass
class RootLogConfig:
    log_format_str: str = '[{engine}] {asctime} - {levelname}: {message} ({name})'
    locale: LocaleType = pkg_locale.getlocale()
    log_to_stdout: bool = True
    log_to_stderr: bool = True
    log_to_file: bool = True
    stdout: TextIO = sys.stdout
    stderr: TextIO = sys.stderr
    stdout_log_min_level: int = logging.INFO
    stderr_log_min_level: int = logging.ERROR
    file_log_min_level: int = logging.WARNING
    file_log_path: str = field(default_factory=_get_log_path)



import asyncio
from multiprocessing.pool import ThreadPool

import httpx


async def download_file_to_memory_async(url: str):
    async with httpx.AsyncClient() as client:
        response = await client.get(url)
        if response.status_code == 200:
            return response.content
        else:
            return None


def run_in_thread(url):
    return asyncio.run(download_file_to_memory_async(url))


def download_file_to_memory(url: str) -> bytes | None:
    with ThreadPool(processes=1) as pool:
        async_result = pool.apply_async(run_in_thread, (url,))
        return async_result.get()



from typing import Generic, Type, TypeAlias

from typing_extensions import TypeVar

from omnipy.data.model import Model

from .typedefs import JsonScalar

#
# Private models
#

# Basic building block models

_JsonBaseT = TypeVar(
    '_JsonBaseT',
    bound='JsonScalar | _JsonListM | _JsonDictM | _JsonAnyListM | _JsonAnyDictM',
    default='JsonScalar')


class _JsonListM(Model[list[_JsonBaseT]], Generic[_JsonBaseT]):
    ...


class _JsonDictM(Model[dict[str, _JsonBaseT]], Generic[_JsonBaseT]):
    ...


# Note: This intermediate level of JSON models is needed for two reasons. 1) as targets for
#       .updateForwardRefs(), as this does not seem to work properly directly on a generic model
#       (e.g. `_JsonListM['_JsonAnyUnion'].updateForwardRefs()`), at least in pydantic v1.
#       But even if this is fixed in pydantic v2, or more probably in python 3.13 with PEP649, the
#       intermediate level is still needed due to the other reason: 2) For consistency in the
#       hierarchy of JSON models, as tested in e.g. `test_json_model_consistency_basic()`. The
#       intermediate level of models seems a good solution to make sure the level of the model
#       hierarchy stays the same for e.g. `JsonModel` and `JsonDictModel`.


class _JsonAnyListM(_JsonListM['_JsonAnyUnion']):
    ...


class _JsonAnyDictM(_JsonDictM['_JsonAnyUnion']):
    ...


class _JsonOnlyListsM(_JsonListM['_JsonOnlyListsUnion']):
    ...


class _JsonOnlyDictsM(_JsonDictM['_JsonOnlyDictsUnion']):
    ...


# TypeAliases
# TODO: Consider whether these TypeAliases are needed in pydantic v2. In v1 they are needed to for
#       the hack for propagating None to work. Removing this level will simplify JSON models.
#       If updated, also update frozen models

_JsonAnyUnion: TypeAlias = JsonScalar | _JsonAnyListM | _JsonAnyDictM
_JsonOnlyListsUnion: TypeAlias = JsonScalar | _JsonOnlyListsM
_JsonOnlyDictsUnion: TypeAlias = JsonScalar | _JsonOnlyDictsM

_JsonListOfScalars: TypeAlias = _JsonListM[JsonScalar]

_JsonDictOfScalars: TypeAlias = _JsonDictM[JsonScalar]

# Basic models needs to update their forward_refs with type aliases declared above

_JsonAnyListM.update_forward_refs()
_JsonAnyDictM.update_forward_refs()
_JsonOnlyListsM.update_forward_refs()
_JsonOnlyDictsM.update_forward_refs()

#
# Exportable models
#

# General


class JsonModel(Model[_JsonAnyUnion]):
    """
    JsonModel is a general JSON model supporting any JSON content, any levels deep.

    Examples:
        >>> my_int_json = JsonModel(123)
        >>> my_int_json.to_data()
        123
        >>> my_int_json.to_json()
        '123'
        >>> my_json = JsonModel([True, {'a': None, 'b': [1, 12.5, 'abc']}])
        >>> my_json.to_data()
        [True, {'a': None, 'b': [1, 12.5, 'abc']}]
        >>> my_json.to_json()
        '[true, {"a": null, "b": [1, 12.5, "abc"]}]'
        >>> print(my_json.to_json(pretty=True))
        [
          true,
          {
            "a": null,
            "b": [
              1,
              12.5,
              "abc"
            ]
          }
        ]
        >>> try:
        ...     my_json = JsonModel([{'a': [1, 2, 1+2j]}])  # nested complex number
        ... except Exception as e:
        ...     print(str(e).splitlines()[0])
        34 validation errors for JsonModel
    """


# Scalars

JsonScalarModel: TypeAlias = Model[JsonScalar]
"""
JsonScalarModel is a limited JSON model supporting only scalar JSON content, e.g. the basic
types: `None`, `int`, `float`, `str`, and `bool`. Lists and dicts (or "objects") are not
supported.

Examples:
    >>> my_none = JsonScalarModel(None)
    >>> my_none.to_data(), my_none.to_json()
    (None, 'null')
    >>> my_int = JsonScalarModel(123)
    >>> my_int.to_data(), my_int.to_json()
    (123, '123')
    >>> my_float = JsonScalarModel(12.3)
    >>> my_float.to_data(), my_float.to_json()
    (12.3, '12.3')
    >>> my_str = JsonScalarModel('abc')
    >>> my_str.to_data(), my_str.to_json()
    (abc, '"abc"')
    >>> my_bool = JsonScalarModel(False)
    >>> my_bool.to_data(), my_bool.to_json()
    (False, '"false"')
    >>> try:
    ...     my_json = JsonScalarModel([123])
    ... except Exception as e:
    ...     print(str(e).splitlines()[0])
    6 validation errors for JsonScalarModel
"""

# List at the top level


class JsonListModel(Model[_JsonAnyListM]):
    """
    JsonListModel is a limited JSON model supporting only JSON content that has a list (or "array"
    in JSON nomenclature) at the root. The contents of the top-level list can be any JSON content,
    though, any levels deep.

    Examples:
        >>> my_json = JsonListModel([True, {'a': None, 'b': [1, 12.5, 'abc']}])
        >>> my_json.to_data()
        [True, {'a': None, 'b': [1, 12.5, 'abc']}]
        >>> my_json.to_json()
        '[true, {"a": null, "b": [1, 12.5, "abc"]}]'
        >>> print(my_json.to_json(pretty=True))
        [
          true,
          {
            "a": null,
            "b": [
              1,
              12.5,
              "abc"
            ]
          }
        ]
        >>> try:
        ...     my_json = JsonListModel({'a': None, 'b': [1, 12.5, {'abc': 123}]})
        ... except Exception as e:
        ...     print(str(e).splitlines()[0])
        3 validation errors for JsonListModel
    """


class JsonListOfScalarsModel(Model[_JsonListOfScalars]):
    ...


class JsonListOfListsModel(Model[_JsonListM[_JsonAnyListM]]):
    ...


class JsonListOfListsOfScalarsModel(Model[_JsonListM[_JsonListOfScalars]]):
    ...


class JsonListOfDictsModel(Model[_JsonListM[_JsonAnyDictM]]):
    ...


class JsonListOfDictsOfScalarsModel(Model[_JsonListM[_JsonDictOfScalars]]):
    ...


# Dict at the top level


class JsonDictModel(Model[_JsonAnyDictM]):
    """
    JsonDictModel is a limited JSON model supporting only JSON content that has a dict (or "object"
    in JSON nomenclature) at the root. The values of the top-level dict can be any JSON content,
    though, any levels deep.

    Examples:
        >>> my_json = JsonDictModel({'a': None, 'b': [1, 12.5, {'abc': 123}]})
        >>> my_json.to_data()
        {'a': None, 'b': [1, 12.5, {'abc': 123}]}
        >>> my_json.to_json()
        '{"a": null, "b": [1, 12.5, {"abc": 123}]}'
        >>> print(my_json.to_json(pretty=True))
        {
          "a": null,
          "b": [
            1,
            12.5,
            {
              "abc": 123
            }
          ]
        }
        >>> try:
        ...     my_json = JsonDictModel([True, {'a': None, 'b': [1, 12.5, 'abc']}])
        ... except Exception as e:
        ...     print(str(e).splitlines()[0])
        3 validation errors for JsonDictModel
    """


class JsonDictOfScalarsModel(Model[_JsonDictOfScalars]):
    ...


class JsonDictOfListsModel(Model[_JsonDictM[_JsonAnyListM]]):
    ...


class JsonDictOfListsOfScalarsModel(Model[_JsonDictM[_JsonListOfScalars]]):
    ...


class JsonDictOfDictsModel(Model[_JsonDictM[_JsonAnyDictM]]):
    ...


class JsonDictOfDictsOfScalarsModel(Model[_JsonDictM[_JsonDictOfScalars]]):
    ...


# Nested models


class JsonOnlyListsModel(Model[_JsonOnlyListsUnion]):
    ...


class JsonNestedListsModel(Model[_JsonOnlyListsM]):
    ...


class JsonOnlyDictsModel(Model[_JsonOnlyDictsUnion]):
    ...


class JsonNestedDictsModel(Model[_JsonOnlyDictsM]):
    ...


# More specific models


class JsonListOfNestedDictsModel(Model[_JsonListM[_JsonOnlyDictsM]]):
    ...


class JsonDictOfNestedListsModel(Model[_JsonDictM[_JsonOnlyListsM]]):
    ...


class JsonDictOfListsOfDictsModel(Model[_JsonDictM[_JsonListM[_JsonAnyDictM]]]):
    ...


# Custom models


class JsonCustomListModel(Model[_JsonListM[_JsonBaseT]], Generic[_JsonBaseT]):
    ...


class JsonCustomDictModel(Model[_JsonDictM[_JsonBaseT]], Generic[_JsonBaseT]):
    ...


# Add note of dict keys to models containing dicts

_NOTE_ON_DICT_KEYS = """
    Note:
        JSON dicts (or "objects") only supports strings as keys. By default, however, omnipy
        JSON models parse the basic types `float`, `int`, and `bool` as strings if used as keys
        in JSON dicts/objects.

        Example:

            >>> my_dict = JsonModel({'a': None, 0.5: 12.3, 1: 123, False: True})
            >>> my_dict.to_data()
            {'a': None, '0.5': 12.3, '1': 123, 'False': True}
            >>> my_dict.to_json()
            '{"a": null, "0.5": 12.3, "1": 123, "False": true}'
    """

_dict_related_json_model_classes: list[Type[Model]] = [
    JsonModel,
    JsonDictModel,
    JsonDictOfScalarsModel,
    JsonDictOfListsModel,
    JsonDictOfListsOfScalarsModel,
    JsonDictOfDictsModel,
    JsonDictOfDictsOfScalarsModel,
    JsonOnlyDictsModel,
    JsonNestedListsModel,
    JsonListOfNestedDictsModel,
    JsonDictOfNestedListsModel,
    JsonDictOfListsOfDictsModel,
    JsonCustomDictModel,
]

for _dict_related_json_model_cls in _dict_related_json_model_classes:
    if _dict_related_json_model_cls.__doc__:
        _dict_related_json_model_cls.__doc__ += _NOTE_ON_DICT_KEYS



from typing import Callable, cast, Generic, TypeAlias

from typing_extensions import TypeVar

from omnipy.data.helpers import obj_or_model_contents_isinstance
from omnipy.data.model import Model
from omnipy.data.param import bind_adjust_model_func, params_dataclass, ParamsBase


class _EncodingParamsMixin:
    @params_dataclass
    class Params(ParamsBase):
        encoding: str = 'utf-8'


class _BytesModel(Model[bytes | str], _EncodingParamsMixin):
    @classmethod
    def _parse_data(cls, data: bytes | str) -> bytes:
        if isinstance(data, bytes):
            return data

        return data.encode(cls.Params.encoding)


class BytesModel(_BytesModel):
    adjust = bind_adjust_model_func(
        _BytesModel.clone_model_cls,
        _EncodingParamsMixin.Params,
    )


class _StrModel(Model[str | bytes], _EncodingParamsMixin):
    Params = _EncodingParamsMixin.Params

    @classmethod
    def _parse_data(cls, data: str | bytes) -> str:
        if isinstance(data, str):
            return data

        return data.decode(cls.Params.encoding)

    ...


class StrModel(_StrModel):
    adjust = bind_adjust_model_func(
        _StrModel.clone_model_cls,
        _EncodingParamsMixin.Params,
    )


class _SplitToItemsMixin:
    @params_dataclass
    class Params(ParamsBase):
        strip: bool = True
        strip_chars: str | None = None
        delimiter: str = ','

    @classmethod
    def _split_line(cls, data: str) -> list[str]:
        strip = cls.Params.strip
        strip_chars = cls.Params.strip_chars
        delimiter = cls.Params.delimiter

        if strip:
            data = data.strip(strip_chars)

        items = data.split(delimiter)
        return [item.strip(strip_chars) for item in items] if strip else items


class _SplitToItemsModel(
        Model[list[str] | str],
        _SplitToItemsMixin,
):
    @classmethod
    def _parse_data(cls, data: list[str] | str) -> list[str]:
        if isinstance(data, list):
            return data

        return cls._split_line(data)


class SplitToItemsModel(_SplitToItemsModel):
    adjust = bind_adjust_model_func(
        _SplitToItemsModel.clone_model_cls,
        _SplitToItemsModel.Params,
    )


class _SplitToLinesParamsMixin:
    @params_dataclass
    class Params(ParamsBase):
        strip: bool = True
        strip_chars: str | None = None
        delimiter: str = '\n'


class SplitToLinesModel(_SplitToLinesParamsMixin, _SplitToItemsModel):
    adjust = bind_adjust_model_func(
        _SplitToItemsModel.clone_model_cls,
        _SplitToLinesParamsMixin.Params,
    )


class _SplitItemsToSubitemsModel(
        Model[list[list[str]] | list[str] | list[StrModel]],
        _SplitToItemsMixin,
):
    @classmethod
    def _parse_data(cls, data: list[list[str]] | list[str] | list[StrModel]) -> list[list[str]]:
        if isinstance(data, list) and (len(data) == 0 or isinstance(data[0], list)):
            return cast(list[list[str]], data)

        return [cls._split_line(cast(str, line)) for line in data]


class SplitItemsToSubitemsModel(_SplitItemsToSubitemsModel):
    adjust = bind_adjust_model_func(
        _SplitItemsToSubitemsModel.clone_model_cls,
        _SplitItemsToSubitemsModel.Params,
    )


class _SplitLinesToColumnsParamsMixin:
    @params_dataclass
    class Params(ParamsBase):
        strip: bool = True
        strip_chars: str | None = None
        delimiter: str = '\t'


class SplitLinesToColumnsModel(_SplitLinesToColumnsParamsMixin, _SplitItemsToSubitemsModel):
    adjust = bind_adjust_model_func(
        _SplitItemsToSubitemsModel.clone_model_cls,
        _SplitLinesToColumnsParamsMixin.Params,
    )


class _JoinItemsMixin:
    @params_dataclass
    class Params(ParamsBase):
        delimiter: str = ','

    @classmethod
    def _join_items(cls, data: list[str]) -> str:
        return cls.Params.delimiter.join(data)


class _JoinItemsModel(Model[str | list[str]], _JoinItemsMixin):
    @classmethod
    def _parse_data(cls, data: str | list[str]) -> str:
        if isinstance(data, str):
            return data

        return cls._join_items(data)

    ...


class JoinItemsModel(_JoinItemsModel):
    adjust = bind_adjust_model_func(
        _JoinItemsModel.clone_model_cls,
        _JoinItemsModel.Params,
    )


class _JoinLinesParamsMixin:
    @params_dataclass
    class Params(ParamsBase):
        delimiter: str = '\n'


class JoinLinesModel(_JoinLinesParamsMixin, _JoinItemsModel):
    adjust = bind_adjust_model_func(
        _JoinItemsModel.clone_model_cls,
        _JoinLinesParamsMixin.Params,
    )


class _JoinSubitemsToItemsModel(
        Model[list[str] | list[list[str]]],
        _JoinItemsMixin,
):
    @classmethod
    def _parse_data(cls, data: list[str] | list[list[str]]) -> list[str]:
        if isinstance(data, list) and (len(data) == 0 or not isinstance(data[0], list)):
            return cast(list[str], data)

        return [cls._join_items(cast(list[str], cols)) for cols in data]


class JoinSubitemsToItemsModel(_JoinSubitemsToItemsModel):
    adjust = bind_adjust_model_func(
        _JoinSubitemsToItemsModel.clone_model_cls,
        _JoinSubitemsToItemsModel.Params,
    )


class _JoinItemsByTabParamsMixin:
    @params_dataclass
    class Params(ParamsBase):
        delimiter: str = ('\t')


class JoinColumnsToLinesModel(_JoinItemsByTabParamsMixin, _JoinSubitemsToItemsModel):
    adjust = bind_adjust_model_func(
        _JoinSubitemsToItemsModel.clone_model_cls,
        _JoinItemsByTabParamsMixin.Params,
    )


NestedStrWithListModelsT = TypeVar(
    'NestedStrWithListModelsT', bound='NestedStrWithListModels', default='str')


class ListOfNestedStrModel(Model[list[NestedStrWithListModelsT]],
                           Generic[NestedStrWithListModelsT]):
    ...


NestedStrWithListModels: TypeAlias = str | ListOfNestedStrModel

ListOfNestedStrModel.update_forward_refs()

NestedStrNoListModelsT = TypeVar(
    'NestedStrNoListModelsT', default=str | list['NestedStrNoListModelsT'])  # type: ignore[misc]


class _NestedItemsParamsMixin:
    @params_dataclass
    class Params(ParamsBase):
        delimiters: tuple[str, ...] = ()

    @classmethod
    def _split_data_according_to_delimiters(cls,
                                            data: NestedStrWithListModels,
                                            level: int = 0) -> str | list[NestedStrNoListModelsT]:

        raw_data = cast(str | list[NestedStrWithListModels],
                        data if isinstance(data, str) else data.contents)

        num_delimiters = len(cls.Params.delimiters)

        if level == 0 and len(raw_data) == 0:
            return [] if num_delimiters > 0 else ''

        split_data: list[NestedStrWithListModels]
        if isinstance(raw_data, str):
            if num_delimiters == 0:
                return raw_data

            split_data = cast(list[NestedStrWithListModels],
                              raw_data.split(cls.Params.delimiters[level]))
        else:
            split_data = raw_data

        next_level = level + 1

        if num_delimiters == next_level:
            assert len(split_data) > 1, \
                (f'Data at bottom level (level {level}) must contain at '
                 f'least one delimiter or equivalently be pre-split into '
                 f'more than one item (bottom delimiter: '
                 f'{cls.Params.delimiters[level]}): {split_data}')
        if num_delimiters > next_level:
            split_data_no_list_models = [
                cls._split_data_according_to_delimiters(item, level=next_level)
                for item in split_data
            ]
            return cast(list[NestedStrNoListModelsT], split_data_no_list_models)
        else:
            if num_delimiters == 0:
                assert isinstance(data, str), \
                    'Data must be an string if no delimiters are provided.'
            else:
                assert not any(obj_or_model_contents_isinstance(item, list) for item in data), \
                    (f'Data is nested higher than permitted by the number of delimiters in '
                     f'Params (={num_delimiters}).')

        return cast(list[NestedStrNoListModelsT], split_data)


class _NestedSplitToItemsModel(Model[list[NestedStrNoListModelsT] | str],
                               Generic[NestedStrNoListModelsT],
                               _NestedItemsParamsMixin):
    @classmethod
    def _parse_data(cls,
                    data: list[NestedStrNoListModelsT] | str) -> list[NestedStrNoListModelsT] | str:
        str_parsed_data = Model[NestedStrWithListModels](data).contents
        return cls._split_data_according_to_delimiters(str_parsed_data)


class NestedSplitToItemsModel(_NestedSplitToItemsModel):
    adjust = bind_adjust_model_func(
        cast(Callable[..., type[_NestedSplitToItemsModel]],
             _NestedSplitToItemsModel.clone_model_cls),
        _NestedSplitToItemsModel.Params,
    )


class _NestedJoinItemsModel(Model[str | list[NestedStrNoListModelsT]],
                            Generic[NestedStrNoListModelsT],
                            _NestedItemsParamsMixin):
    @classmethod
    def _join_data_according_to_delimiters(cls,
                                           data: str | list[NestedStrNoListModelsT],
                                           level: int = 0) -> str:
        if isinstance(data, str):
            return data

        num_delimiters = len(cls.Params.delimiters)
        next_level = level + 1
        raw_data: list[str]
        if num_delimiters > next_level:
            raw_data = [
                cls._join_data_according_to_delimiters(
                    cast(str | list[NestedStrNoListModelsT], item), level=next_level)
                for item in data
            ]
        else:
            raw_data = cast(list[str], data)

        return cls.Params.delimiters[level].join(raw_data) if len(raw_data) > 0 else ''

    @classmethod
    def _parse_data(cls, data: str | list[NestedStrNoListModelsT]) -> str:
        str_parsed_data = Model[NestedStrWithListModels](data).contents
        return cls._join_data_according_to_delimiters(
            cls._split_data_according_to_delimiters(str_parsed_data))


class NestedJoinItemsModel(_NestedJoinItemsModel):
    adjust = bind_adjust_model_func(
        cast(Callable[..., type[_NestedJoinItemsModel]], _NestedJoinItemsModel.clone_model_cls),
        _NestedItemsParamsMixin.Params,
    )


# TODO: Make MatchItemsModel Generic. Generics are currently not supported by Parametrized Models.


class _MatchItemsModel(Model[list[str]]):
    @params_dataclass
    class Params(ParamsBase):
        match_functions: tuple[Callable[[str], bool], ...] = ()
        invert_matches: bool = False
        match_all: bool = True

    @classmethod
    def _parse_data(cls, data: list[str]) -> list[str]:
        match_functions = cls.Params.match_functions
        invert_matches = cls.Params.invert_matches

        if len(match_functions) == 0:
            return data

        logic_operator = all if cls.Params.match_all else any

        return [
            item for item in data
            if (logic_operator(match_func(item) for match_func in match_functions) is True)
            ^ invert_matches
        ]


class MatchItemsModel(_MatchItemsModel):
    adjust = bind_adjust_model_func(
        _MatchItemsModel.clone_model_cls,
        _MatchItemsModel.Params,
    )
